#!/usr/bin/env python

"""
<Program Name>
  keys.py

<Author>
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  October 4, 2013.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  The goal of this module is to centralize cryptographic key routines and their
  supported operations (e.g., creating and verifying signatures).  This module
  is designed to support multiple public-key algorithms, such as RSA, Ed25519,
  and ECDSA, and multiple cryptography libraries.  Which cryptography library
  to use is determined by the default, or user modified, values set in
  'settings.py'

  https://en.wikipedia.org/wiki/RSA_(algorithm)
  http://ed25519.cr.yp.to/
  https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm

  The (RSA, ECDSA and Ed25519)-related functions provided include
  generate_rsa_key(), generate_ed25519_key(), generate_ecdsa_key(),
  create_signature(), and verify_signature().  The cryptography libraries
  called by 'securesystemslib.keys.py' generate the actual keys and the
  functions listed above can be viewed as the easy-to-use public interface.

  Additional functions contained here include format_keyval_to_metadata() and
  format_metadata_to_key().  These last two functions produce or use keys
  compatible with the key structures listed in Metadata files.  The key
  generation functions return a dictionary containing all the information needed
  of keys, such as public & private keys, and a keyID.  create_signature()
  and verify_signature() are supplemental functions needed for generating
  signatures and verifying them.

  Key IDs are used as identifiers for keys (e.g., RSA key).  They are the
  hexadecimal representation of the hash of the key object (specifically, the
  key object containing only the public key).  Review the '_get_keyid()'
  function of this module to see precisely how keyids are generated.  One may
  get the keyid of a key object by simply accessing the dictionary's 'keyid'
  key (i.e., rsakey['keyid']).
 """

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

# Required for hexadecimal conversions.  Signatures and public/private keys are
# hexlified.
import binascii

import logging

from securesystemslib import ecdsa_keys
from securesystemslib import ed25519_keys
from securesystemslib import exceptions
from securesystemslib import formats
from securesystemslib import rsa_keys
from securesystemslib import settings
from securesystemslib import util
from securesystemslib.hash import digest


# The hash algorithm to use in the generation of keyids.
_KEY_ID_HASH_ALGORITHM = 'sha256'

# Recommended RSA key sizes:
# http://www.emc.com/emc-plus/rsa-labs/historical/twirl-and-rsa-key-size.htm#table1
# According to the document above, revised May 6, 2003, RSA keys of
# size 3072 provide security through 2031 and beyond.
_DEFAULT_RSA_KEY_BITS = 3072


RSA_SIGNATURE_SCHEMES = [
  'rsassa-pss-md5',
  'rsassa-pss-sha1',
  'rsassa-pss-sha224',
  'rsassa-pss-sha256',
  'rsassa-pss-sha384',
  'rsassa-pss-sha512',
  'rsa-pkcs1v15-md5',
  'rsa-pkcs1v15-sha1',
  'rsa-pkcs1v15-sha224',
  'rsa-pkcs1v15-sha256',
  'rsa-pkcs1v15-sha384',
  'rsa-pkcs1v15-sha512',
]

logger = logging.getLogger(__name__)


def generate_rsa_key(bits=_DEFAULT_RSA_KEY_BITS, scheme='rsassa-pss-sha256'):
  """
  <Purpose>
    Generate public and private RSA keys, with modulus length 'bits'.  In
    addition, a keyid identifier for the RSA key is generated.  The object
    returned conforms to 'securesystemslib.formats.RSAKEY_SCHEMA' and has the
    form:

    {'keytype': 'rsa',
     'scheme': 'rsassa-pss-sha256',
     'keyid': keyid,
     'keyval': {'public': '-----BEGIN RSA PUBLIC KEY----- ...',
                'private': '-----BEGIN RSA PRIVATE KEY----- ...'}}

    The public and private keys are strings in PEM format.

    Although the PyCA cryptography library and/or its crypto backend might set
    a minimum key size, generate() enforces a minimum key size of 2048 bits.
    If 'bits' is unspecified, a 3072-bit RSA key is generated, which is the key
    size recommended by securesystemslib.  These key size restrictions are only
    enforced for keys generated within securesystemslib.  RSA keys with sizes
    lower than what we recommended may still be imported (e.g., with
    import_rsakey_from_pem().

    >>> rsa_key = generate_rsa_key(bits=2048)
    >>> securesystemslib.formats.RSAKEY_SCHEMA.matches(rsa_key)
    True

    >>> public = rsa_key['keyval']['public']
    >>> private = rsa_key['keyval']['private']
    >>> securesystemslib.formats.PEMRSA_SCHEMA.matches(public)
    True
    >>> securesystemslib.formats.PEMRSA_SCHEMA.matches(private)
    True

  <Arguments>
    bits:
      The key size, or key length, of the RSA key.  'bits' must be 2048, or
      greater, and a multiple of 256.

    scheme:
      The signature scheme used by the key.  It must be one from the list
      `securesystemslib.keys.RSA_SIGNATURE_SCHEMES`.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'bits' is improperly or invalid
    (i.e., not an integer and not at least 2048).

    ValueError, if an exception occurs after calling the RSA key generation
    routine.  The 'ValueError' exception is raised by the key generation
    function of the cryptography library called.

  <Side Effects>
    None.

  <Returns>
    A dictionary containing the RSA keys and other identifying information.
    Conforms to 'securesystemslib.formats.RSAKEY_SCHEMA'.
  """

  # Does 'bits' have the correct format?  This check will ensure 'bits'
  # conforms to 'securesystemslib.formats.RSAKEYBITS_SCHEMA'.  'bits' must be
  # an integer object, with a minimum value of 2048.  Raise
  # 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.RSAKEYBITS_SCHEMA.check_match(bits)
  formats.RSA_SCHEME_SCHEMA.check_match(scheme)

  # Begin building the RSA key dictionary.
  rsakey_dict = {}
  keytype = 'rsa'
  public = None
  private = None

  # Generate the public and private RSA keys.  The pyca/cryptography module is
  # used to generate the actual key.  Raise 'ValueError' if 'bits' is less than
  # 1024, although a 2048-bit minimum is enforced by
  # securesystemslib.formats.RSAKEYBITS_SCHEMA.check_match().
  public, private = rsa_keys.generate_rsa_public_and_private(bits)

  # When loading in PEM keys, extract_pem() is called, which strips any
  # leading or trailing new line characters. Do the same here before generating
  # the keyid.
  public =  extract_pem(public, private_pem=False)
  private = extract_pem(private, private_pem=True)

  # Generate the keyid of the RSA key.  Note: The private key material is not
  # included in the generation of the 'keyid' identifier.  Convert any '\r\n'
  # (e.g., Windows) newline characters to '\n' so that a consistent keyid is
  # generated.
  key_value = {'public': public.replace('\r\n', '\n'),
               'private': ''}
  keyid = _get_keyid(keytype, scheme, key_value)

  # Build the 'rsakey_dict' dictionary.  Update 'key_value' with the RSA
  # private key prior to adding 'key_value' to 'rsakey_dict'.
  key_value['private'] = private

  rsakey_dict['keytype'] = keytype
  rsakey_dict['scheme'] = scheme
  rsakey_dict['keyid'] = keyid
  rsakey_dict['keyid_hash_algorithms'] = settings.HASH_ALGORITHMS
  rsakey_dict['keyval'] = key_value

  return rsakey_dict





def generate_ecdsa_key(scheme='ecdsa-sha2-nistp256'):
  """
  <Purpose>
    Generate public and private ECDSA keys, with NIST P-256 + SHA256 (for
    hashing) being the default scheme.  In addition, a keyid identifier for the
    ECDSA key is generated.  The object returned conforms to
    'securesystemslib.formats.ECDSAKEY_SCHEMA' and has the form:

    {'keytype': 'ecdsa',
     'scheme', 'ecdsa-sha2-nistp256',
     'keyid': keyid,
     'keyval': {'public': '',
                'private': ''}}

    The public and private keys are strings in TODO format.

    >>> ecdsa_key = generate_ecdsa_key(scheme='ecdsa-sha2-nistp256')
    >>> securesystemslib.formats.ECDSAKEY_SCHEMA.matches(ecdsa_key)
    True

  <Arguments>
    scheme:
      The ECDSA signature scheme.  By default, ECDSA NIST P-256 is used, with
      SHA256 for hashing.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'scheme' is improperly
    formatted or invalid (i.e., not one of the supported ECDSA signature
    schemes).

  <Side Effects>
    None.

  <Returns>
    A dictionary containing the ECDSA keys and other identifying information.
    Conforms to 'securesystemslib.formats.ECDSAKEY_SCHEMA'.
  """

  # Does 'scheme' have the correct format?
  # This check will ensure 'scheme' is properly formatted and is a supported
  # ECDSA signature scheme.  Raise 'securesystemslib.exceptions.FormatError' if
  # the check fails.
  formats.ECDSA_SCHEME_SCHEMA.check_match(scheme)

  # Begin building the ECDSA key dictionary.
  ecdsa_key = {}
  keytype = 'ecdsa'
  public = None
  private = None

  # Generate the public and private ECDSA keys with one of the supported
  # libraries.
  public, private = ecdsa_keys.generate_public_and_private(scheme)

  # Generate the keyid of the Ed25519 key.  'key_value' corresponds to the
  # 'keyval' entry of the 'Ed25519KEY_SCHEMA' dictionary.  The private key
  # information is not included in the generation of the 'keyid' identifier.
  # Convert any '\r\n' (e.g., Windows) newline characters to '\n' so that a
  # consistent keyid is generated.
  key_value = {'public': public.replace('\r\n', '\n'),
               'private': ''}
  keyid = _get_keyid(keytype, scheme, key_value)

  # Build the 'ed25519_key' dictionary.  Update 'key_value' with the Ed25519
  # private key prior to adding 'key_value' to 'ed25519_key'.

  key_value['private'] = private

  ecdsa_key['keytype'] = keytype
  ecdsa_key['scheme'] = scheme
  ecdsa_key['keyid'] = keyid
  ecdsa_key['keyval'] = key_value

  # Add "keyid_hash_algorithms" so that equal ECDSA keys with different keyids
  # can be associated using supported keyid_hash_algorithms.
  ecdsa_key['keyid_hash_algorithms'] = settings.HASH_ALGORITHMS

  return ecdsa_key





def generate_ed25519_key(scheme='ed25519'):
  """
  <Purpose>
    Generate public and private ED25519 keys, both of length 32-bytes, although
    they are hexlified to 64 bytes.  In addition, a keyid identifier generated
    for the returned ED25519 object.  The object returned conforms to
    'securesystemslib.formats.ED25519KEY_SCHEMA' and has the form:

    {'keytype': 'ed25519',
     'scheme': 'ed25519',
     'keyid': 'f30a0870d026980100c0573bd557394f8c1bbd6...',
     'keyval': {'public': '9ccf3f02b17f82febf5dd3bab878b767d8408...',
                'private': 'ab310eae0e229a0eceee3947b6e0205dfab3...'}}

    >>> ed25519_key = generate_ed25519_key()
    >>> securesystemslib.formats.ED25519KEY_SCHEMA.matches(ed25519_key)
    True
    >>> len(ed25519_key['keyval']['public'])
    64
    >>> len(ed25519_key['keyval']['private'])
    64

  <Arguments>
    scheme:
      The signature scheme used by the generated Ed25519 key.

  <Exceptions>
    None.

  <Side Effects>
    The ED25519 keys are generated by calling either the optimized pure Python
    implementation of ed25519, or the ed25519 routines provided by 'pynacl'.

  <Returns>
    A dictionary containing the ED25519 keys and other identifying information.
    Conforms to 'securesystemslib.formats.ED25519KEY_SCHEMA'.
  """

  # Are the arguments properly formatted?  If not, raise an
  # 'securesystemslib.exceptions.FormatError' exceptions.
  formats.ED25519_SIG_SCHEMA.check_match(scheme)

  # Begin building the Ed25519 key dictionary.
  ed25519_key = {}
  keytype = 'ed25519'
  public = None
  private = None

  # Generate the public and private Ed25519 key with the 'pynacl' library.
  # Unlike in the verification of Ed25519 signatures, do not fall back to the
  # optimized, pure python implementation provided by PyCA.  Ed25519 should
  # always be generated with a backend like libsodium to prevent side-channel
  # attacks.
  public, private = ed25519_keys.generate_public_and_private()

  # Generate the keyid of the Ed25519 key.  'key_value' corresponds to the
  # 'keyval' entry of the 'Ed25519KEY_SCHEMA' dictionary.  The private key
  # information is not included in the generation of the 'keyid' identifier.
  key_value = {'public': binascii.hexlify(public).decode(),
               'private': ''}
  keyid = _get_keyid(keytype, scheme, key_value)

  # Build the 'ed25519_key' dictionary.  Update 'key_value' with the Ed25519
  # private key prior to adding 'key_value' to 'ed25519_key'.
  key_value['private'] = binascii.hexlify(private).decode()

  ed25519_key['keytype'] = keytype
  ed25519_key['scheme'] = scheme
  ed25519_key['keyid'] = keyid
  ed25519_key['keyid_hash_algorithms'] = settings.HASH_ALGORITHMS
  ed25519_key['keyval'] = key_value

  return ed25519_key





def format_keyval_to_metadata(keytype, scheme, key_value, private=False):
  """
  <Purpose>
    Return a dictionary conformant to 'securesystemslib.formats.KEY_SCHEMA'.
    If 'private' is True, include the private key.  The dictionary
    returned has the form:

    {'keytype': keytype,
     'scheme' : scheme,
     'keyval': {'public': '...',
                'private': '...'}}

    or if 'private' is False:

    {'keytype': keytype,
     'scheme': scheme,
     'keyval': {'public': '...',
                'private': ''}}

    >>> ed25519_key = generate_ed25519_key()
    >>> key_val = ed25519_key['keyval']
    >>> keytype = ed25519_key['keytype']
    >>> scheme = ed25519_key['scheme']
    >>> ed25519_metadata = \
    format_keyval_to_metadata(keytype, scheme, key_val, private=True)
    >>> securesystemslib.formats.KEY_SCHEMA.matches(ed25519_metadata)
    True

  <Arguments>
    key_type:
      The 'rsa' or 'ed25519' strings.

    scheme:
      The signature scheme used by the key.

    key_value:
      A dictionary containing a private and public keys.
      'key_value' is of the form:

      {'public': '...',
       'private': '...'}},

      conformant to 'securesystemslib.formats.KEYVAL_SCHEMA'.

    private:
      Indicates if the private key should be included in the dictionary
      returned.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'key_value' does not conform to
    'securesystemslib.formats.KEYVAL_SCHEMA', or if the private key is not
    present in 'key_value' if requested by the caller via 'private'.

  <Side Effects>
    None.

  <Returns>
    A 'securesystemslib.formats.KEY_SCHEMA' dictionary.
  """

  # Does 'keytype' have the correct format?
  # This check will ensure 'keytype' has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.KEYTYPE_SCHEMA.check_match(keytype)

  # Does 'scheme' have the correct format?
  formats.SCHEME_SCHEMA.check_match(scheme)

  # Does 'key_value' have the correct format?
  formats.KEYVAL_SCHEMA.check_match(key_value)

  if private is True:
    # If the caller requests (via the 'private' argument) to include a private
    # key in the returned dictionary, ensure the private key is actually
    # present in 'key_val' (a private key is optional for 'KEYVAL_SCHEMA'
    # dicts).
    if 'private' not in key_value:
      raise exceptions.FormatError('The required private key'
        ' is missing from: ' + repr(key_value))

    else:
      return {'keytype': keytype, 'scheme': scheme, 'keyval': key_value}

  else:
    public_key_value = {'public': key_value['public']}

    return {'keytype': keytype,
            'scheme': scheme,
            'keyid_hash_algorithms': settings.HASH_ALGORITHMS,
            'keyval': public_key_value}





def format_metadata_to_key(key_metadata, default_keyid=None,
    keyid_hash_algorithms=None):
  """
  <Purpose>
    Construct a key dictionary (e.g., securesystemslib.formats.RSAKEY_SCHEMA)
    according to the keytype of 'key_metadata'.  The dict returned by this
    function has the exact format as the dict returned by one of the key
    generations functions, like generate_ed25519_key().  The dict returned
    has the form:

    {'keytype': keytype,
     'scheme': scheme,
     'keyid': 'f30a0870d026980100c0573bd557394f8c1bbd6...',
     'keyval': {'public': '...',
                'private': '...'}}

    For example, RSA key dictionaries in RSAKEY_SCHEMA format should be used by
    modules storing a collection of keys, such as with keydb.py.  RSA keys as
    stored in metadata files use a different format, so this function should be
    called if an RSA key is extracted from one of these metadata files and need
    converting.  The key generation functions create an entirely new key and
    return it in the format appropriate for 'keydb.py'.

    >>> ed25519_key = generate_ed25519_key()
    >>> key_val = ed25519_key['keyval']
    >>> keytype = ed25519_key['keytype']
    >>> scheme = ed25519_key['scheme']
    >>> ed25519_metadata = \
    format_keyval_to_metadata(keytype, scheme, key_val, private=True)
    >>> ed25519_key_2, junk = format_metadata_to_key(ed25519_metadata)
    >>> securesystemslib.formats.ED25519KEY_SCHEMA.matches(ed25519_key_2)
    True
    >>> ed25519_key == ed25519_key_2
    True

  <Arguments>
    key_metadata:
      The key dictionary as stored in Metadata files, conforming to
      'securesystemslib.formats.KEY_SCHEMA'.  It has the form:

      {'keytype': '...',
       'scheme': scheme,
       'keyval': {'public': '...',
                  'private': '...'}}
    default_keyid:
      A default keyid associated with the key metadata. If this is not
      provided, the keyid will be calculated by _get_keyid using the default
      hash algorithm. If provided, the default keyid can be any string.

    keyid_hash_algorithms:
      An optional list of hash algorithms to use when generating keyids.
      Defaults to securesystemslib.settings.HASH_ALGORITHMS.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'key_metadata' does not conform
    to 'securesystemslib.formats.KEY_SCHEMA'.

  <Side Effects>
    None.

  <Returns>
    In the case of an RSA key, a dictionary conformant to
    'securesystemslib.formats.RSAKEY_SCHEMA'.
  """

  # Does 'key_metadata' have the correct format?
  # This check will ensure 'key_metadata' has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.KEY_SCHEMA.check_match(key_metadata)

  # Construct the dictionary to be returned.
  key_dict = {}
  keytype = key_metadata['keytype']
  scheme = key_metadata['scheme']
  key_value = key_metadata['keyval']

  # Convert 'key_value' to 'securesystemslib.formats.KEY_SCHEMA' and generate
  # its hash The hash is in hexdigest form.
  if default_keyid is None:
    default_keyid = _get_keyid(keytype, scheme, key_value)
  keyids = set()
  keyids.add(default_keyid)

  if keyid_hash_algorithms is None:
    keyid_hash_algorithms = settings.HASH_ALGORITHMS

  for hash_algorithm in keyid_hash_algorithms:
    keyid = _get_keyid(keytype, scheme, key_value, hash_algorithm)
    keyids.add(keyid)

  # All the required key values gathered.  Build 'key_dict'.
  # 'keyid_hash_algorithms'
  key_dict['keytype'] = keytype
  key_dict['scheme'] = scheme
  key_dict['keyid'] = default_keyid
  key_dict['keyid_hash_algorithms'] = keyid_hash_algorithms
  key_dict['keyval'] = key_value

  return key_dict, keyids



def _get_keyid(keytype, scheme, key_value, hash_algorithm = 'sha256'):
  """Return the keyid of 'key_value'."""

  # 'keyid' will be generated from an object conformant to KEY_SCHEMA,
  # which is the format Metadata files (e.g., root.json) store keys.
  # 'format_keyval_to_metadata()' returns the object needed by _get_keyid().
  key_meta = format_keyval_to_metadata(keytype, scheme, key_value, private=False)

  # Convert the key to JSON Canonical format, suitable for adding
  # to digest objects.
  key_update_data = formats.encode_canonical(key_meta)

  # Create a digest object and call update(), using the JSON
  # canonical format of 'rskey_meta' as the update data.
  digest_object = digest(hash_algorithm)
  digest_object.update(key_update_data.encode('utf-8'))

  # 'keyid' becomes the hexadecimal representation of the hash.
  keyid = digest_object.hexdigest()

  return keyid





def create_signature(key_dict, data):
  """
  <Purpose>
    Return a signature dictionary of the form:
    {'keyid': 'f30a0870d026980100c0573bd557394f8c1bbd6...',
     'sig': '...'}.

    The signing process will use the private key in
    key_dict['keyval']['private'] and 'data' to generate the signature.

    The following signature schemes are supported:

    'RSASSA-PSS'
    RFC3447 - RSASSA-PSS
    http://www.ietf.org/rfc/rfc3447.

    'ed25519'
    ed25519 - high-speed high security signatures
    http://ed25519.cr.yp.to/

    Which signature to generate is determined by the key type of 'key_dict'
    and the available cryptography library specified in 'settings'.

    >>> ed25519_key = generate_ed25519_key()
    >>> data = 'The quick brown fox jumps over the lazy dog'
    >>> signature = create_signature(ed25519_key, data)
    >>> securesystemslib.formats.SIGNATURE_SCHEMA.matches(signature)
    True
    >>> len(signature['sig'])
    128
    >>> rsa_key = generate_rsa_key(2048)
    >>> signature = create_signature(rsa_key, data)
    >>> securesystemslib.formats.SIGNATURE_SCHEMA.matches(signature)
    True
    >>> ecdsa_key = generate_ecdsa_key()
    >>> signature = create_signature(ecdsa_key, data)
    >>> securesystemslib.formats.SIGNATURE_SCHEMA.matches(signature)
    True

  <Arguments>
    key_dict:
      A dictionary containing the keys.  An example RSA key dict has the
      form:

      {'keytype': 'rsa',
       'scheme': 'rsassa-pss-sha256',
       'keyid': 'f30a0870d026980100c0573bd557394f8c1bbd6...',
       'keyval': {'public': '-----BEGIN RSA PUBLIC KEY----- ...',
                  'private': '-----BEGIN RSA PRIVATE KEY----- ...'}}

      The public and private keys are strings in PEM format.

    data:
      Data to be signed. This should be a bytes object; data should be
      encoded/serialized before it is passed here.  The same value can be be
      passed into securesystemslib.verify_signature() (along with the public
      key) to later verify the signature.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'key_dict' is improperly
    formatted.

    securesystemslib.exceptions.UnsupportedAlgorithmError, if 'key_dict'
    specifies an unsupported key type or signing scheme.

    TypeError, if 'key_dict' contains an invalid keytype.

  <Side Effects>
    The cryptography library specified in 'settings' is called to perform the
    actual signing routine.

  <Returns>
    A signature dictionary conformant to
    'securesystemslib_format.SIGNATURE_SCHEMA'.
  """

  # Does 'key_dict' have the correct format?
  # This check will ensure 'key_dict' has the appropriate number of objects
  # and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  # The key type of 'key_dict' must be either 'rsa' or 'ed25519'.
  formats.ANYKEY_SCHEMA.check_match(key_dict)

  # Signing the 'data' object requires a private key. Signing schemes that are
  # currently supported are: 'ed25519', 'ecdsa-sha2-nistp256',
  # 'ecdsa-sha2-nistp384' and rsa schemes defined in
  # `securesystemslib.keys.RSA_SIGNATURE_SCHEMES`.
  # RSASSA-PSS and RSA-PKCS1v15 keys and signatures can be generated and
  # verified by rsa_keys.py, and Ed25519 keys by PyNaCl and PyCA's
  # optimized, pure python implementation of Ed25519.
  signature = {}
  keytype = key_dict['keytype']
  scheme = key_dict['scheme']
  public = key_dict['keyval']['public']
  private = key_dict['keyval']['private']
  keyid = key_dict['keyid']
  sig = None

  if keytype == 'rsa':
    if scheme in RSA_SIGNATURE_SCHEMES:
      private = private.replace('\r\n', '\n')
      sig, scheme = rsa_keys.create_rsa_signature(private, data, scheme)

    else:
      raise exceptions.UnsupportedAlgorithmError('Unsupported'
        ' RSA signature scheme specified: ' + repr(scheme))

  elif keytype == 'ed25519':
    public = binascii.unhexlify(public.encode('utf-8'))
    private = binascii.unhexlify(private.encode('utf-8'))
    sig, scheme = ed25519_keys.create_signature(public, private, data, scheme)

  # Continue to support keytypes of ecdsa-sha2-nistp256 and ecdsa-sha2-nistp384
  # for backwards compatibility with older securesystemslib releases
  elif keytype in ['ecdsa', 'ecdsa-sha2-nistp256', 'ecdsa-sha2-nistp384']:
    sig, scheme = ecdsa_keys.create_signature(public, private, data, scheme)

  # 'securesystemslib.formats.ANYKEY_SCHEMA' should have detected invalid key
  # types.  This is a defensive check against an invalid key type.
  else: # pragma: no cover
    raise TypeError('Invalid key type.')

  # Build the signature dictionary to be returned.
  # The hexadecimal representation of 'sig' is stored in the signature.
  signature['keyid'] = keyid
  signature['sig'] = binascii.hexlify(sig).decode()

  return signature





def verify_signature(key_dict, signature, data):
  """
  <Purpose>
    Determine whether the private key belonging to 'key_dict' produced
    'signature'.  verify_signature() will use the public key found in
    'key_dict', the 'sig' objects contained in 'signature', and 'data' to
    complete the verification.

    >>> ed25519_key = generate_ed25519_key()
    >>> data = 'The quick brown fox jumps over the lazy dog'
    >>> signature = create_signature(ed25519_key, data)
    >>> verify_signature(ed25519_key, signature, data)
    True
    >>> verify_signature(ed25519_key, signature, 'bad_data')
    False
    >>> rsa_key = generate_rsa_key()
    >>> signature = create_signature(rsa_key, data)
    >>> verify_signature(rsa_key, signature, data)
    True
    >>> verify_signature(rsa_key, signature, 'bad_data')
    False
    >>> ecdsa_key = generate_ecdsa_key()
    >>> signature = create_signature(ecdsa_key, data)
    >>> verify_signature(ecdsa_key, signature, data)
    True
    >>> verify_signature(ecdsa_key, signature, 'bad_data')
    False

  <Arguments>
    key_dict:
      A dictionary containing the keys and other identifying information.
      If 'key_dict' is an RSA key, it has the form:

      {'keytype': 'rsa',
       'scheme': 'rsassa-pss-sha256',
       'keyid': 'f30a0870d026980100c0573bd557394f8c1bbd6...',
       'keyval': {'public': '-----BEGIN RSA PUBLIC KEY----- ...',
                  'private': '-----BEGIN RSA PRIVATE KEY----- ...'}}

      The public and private keys are strings in PEM format.

    signature:
      The signature dictionary produced by one of the key generation functions.
      'signature' has the form:

      {'keyid': 'f30a0870d026980100c0573bd557394f8c1bbd6...',
       'sig': sig}.

      Conformant to 'securesystemslib.formats.SIGNATURE_SCHEMA'.

    data:
      Data that the signature is expected to be over.  This should be a bytes
      object; data should be encoded/serialized before it is passed here.)
      This is the same value that can be passed into
      securesystemslib.create_signature() in order to create the signature.

  <Exceptions>
    securesystemslib.exceptions.FormatError, raised if either 'key_dict' or
    'signature' are improperly formatted.

    securesystemslib.exceptions.UnsupportedAlgorithmError, if 'key_dict' or
    'signature' specifies an unsupported algorithm.

    securesystemslib.exceptions.CryptoError, if the KEYID in the given
    'key_dict' does not match the KEYID in 'signature'.

  <Side Effects>
    The cryptography library specified in 'settings' called to do the actual
    verification.

  <Returns>
    Boolean.  True if the signature is valid, False otherwise.
  """

  # Does 'key_dict' have the correct format?
  # This check will ensure 'key_dict' has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.ANYKEY_SCHEMA.check_match(key_dict)

  # Does 'signature' have the correct format?
  formats.SIGNATURE_SCHEMA.check_match(signature)

  # Verify that the KEYID in 'key_dict' matches the KEYID listed in the
  # 'signature'.
  if key_dict['keyid'] != signature['keyid']:
    raise exceptions.CryptoError('The KEYID ('
        ' ' + repr(key_dict['keyid']) + ' ) in the given key does not match'
        ' the KEYID ( ' + repr(signature['keyid']) + ' ) in the signature.')

  else:
    logger.debug('The KEYIDs of key_dict and the signature match.')

  # Using the public key belonging to 'key_dict'
  # (i.e., rsakey_dict['keyval']['public']), verify whether 'signature'
  # was produced by key_dict's corresponding private key
  # key_dict['keyval']['private'].
  sig = signature['sig']
  sig = binascii.unhexlify(sig.encode('utf-8'))
  public = key_dict['keyval']['public']
  keytype = key_dict['keytype']
  scheme = key_dict['scheme']
  valid_signature = False


  if keytype == 'rsa':
    if scheme in RSA_SIGNATURE_SCHEMES:
      valid_signature = rsa_keys.verify_rsa_signature(sig,
        scheme, public, data)

    else:
      raise exceptions.UnsupportedAlgorithmError('Unsupported'
          ' signature scheme is specified: ' + repr(scheme))

  elif keytype == 'ed25519':
    if scheme == 'ed25519':
      public = binascii.unhexlify(public.encode('utf-8'))
      valid_signature = ed25519_keys.verify_signature(public,
          scheme, sig, data)

    else:
      raise exceptions.UnsupportedAlgorithmError('Unsupported'
          ' signature scheme is specified: ' + repr(scheme))

  elif keytype in ['ecdsa', 'ecdsa-sha2-nistp256', 'ecdsa-sha2-nistp384']:
    if scheme in ['ecdsa-sha2-nistp256', 'ecdsa-sha2-nistp384']:
      valid_signature = ecdsa_keys.verify_signature(public, scheme, sig, data)

    else:
      raise exceptions.UnsupportedAlgorithmError('Unsupported'
          ' signature scheme is specified: ' + repr(scheme))

  # 'securesystemslib.formats.ANYKEY_SCHEMA' should have detected invalid key
  # types.  This is a defensive check against an invalid key type.
  else: # pragma: no cover
    raise TypeError('Unsupported key type.')

  return valid_signature





def import_rsakey_from_private_pem(pem, scheme='rsassa-pss-sha256', password=None):
  """
  <Purpose>
    Import the private RSA key stored in 'pem', and generate its public key
    (which will also be included in the returned rsakey object).  In addition,
    a keyid identifier for the RSA key is generated.  The object returned
    conforms to 'securesystemslib.formats.RSAKEY_SCHEMA' and has the form:

    {'keytype': 'rsa',
     'scheme': 'rsassa-pss-sha256',
     'keyid': keyid,
     'keyval': {'public': '-----BEGIN RSA PUBLIC KEY----- ...',
                'private': '-----BEGIN RSA PRIVATE KEY----- ...'}}

    The private key is a string in PEM format.

    >>> rsa_key = generate_rsa_key()
    >>> scheme = rsa_key['scheme']
    >>> private = rsa_key['keyval']['private']
    >>> passphrase = 'secret'
    >>> encrypted_pem = create_rsa_encrypted_pem(private, passphrase)
    >>> rsa_key2 = import_rsakey_from_private_pem(encrypted_pem, scheme, passphrase)
    >>> securesystemslib.formats.RSAKEY_SCHEMA.matches(rsa_key)
    True
    >>> securesystemslib.formats.RSAKEY_SCHEMA.matches(rsa_key2)
    True

  <Arguments>
    pem:
      A string in PEM format.  The private key is extracted and returned in
      an rsakey object.

    scheme:
      The signature scheme used by the imported key.

    password: (optional)
      The password, or passphrase, to decrypt the private part of the RSA key
      if it is encrypted.  'password' is not used directly as the encryption
      key, a stronger encryption key is derived from it.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.UnsupportedAlgorithmError, if 'pem' specifies
    an unsupported key type.

  <Side Effects>
    None.

  <Returns>
    A dictionary containing the RSA keys and other identifying information.
    Conforms to 'securesystemslib.formats.RSAKEY_SCHEMA'.
  """

  # Does 'pem' have the correct format?
  # This check will ensure 'pem' conforms to
  # 'securesystemslib.formats.PEMRSA_SCHEMA'.
  formats.PEMRSA_SCHEMA.check_match(pem)

  # Is 'scheme' properly formatted?
  formats.RSA_SCHEME_SCHEMA.check_match(scheme)

  if password is not None:
    formats.PASSWORD_SCHEMA.check_match(password)

  else:
    logger.debug('The password/passphrase is unset.  The PEM is expected'
      ' to be unencrypted.')

  # Begin building the RSA key dictionary.
  rsakey_dict = {}
  keytype = 'rsa'
  public = None
  private = None

  # Generate the public and private RSA keys.  The pyca/cryptography library
  # performs the actual crypto operations.
  public, private = \
    rsa_keys.create_rsa_public_and_private_from_pem(pem, password)

  public =  extract_pem(public, private_pem=False)
  private = extract_pem(private, private_pem=True)

  # Generate the keyid of the RSA key.  'key_value' corresponds to the
  # 'keyval' entry of the 'RSAKEY_SCHEMA' dictionary.  The private key
  # information is not included in the generation of the 'keyid' identifier.
  # Convert any '\r\n' (e.g., Windows) newline characters to '\n' so that a
  # consistent keyid is generated.
  key_value = {'public': public.replace('\r\n', '\n'),
               'private': ''}
  keyid = _get_keyid(keytype, scheme, key_value)

  # Build the 'rsakey_dict' dictionary.  Update 'key_value' with the RSA
  # private key prior to adding 'key_value' to 'rsakey_dict'.
  key_value['private'] = private

  rsakey_dict['keytype'] = keytype
  rsakey_dict['scheme'] = scheme
  rsakey_dict['keyid'] = keyid
  rsakey_dict['keyval'] = key_value

  return rsakey_dict





def import_rsakey_from_public_pem(pem, scheme='rsassa-pss-sha256'):
  """
  <Purpose>
    Generate an RSA key object from 'pem'.  In addition, a keyid identifier for
    the RSA key is generated.  The object returned conforms to
    'securesystemslib.formats.RSAKEY_SCHEMA' and has the form:

    {'keytype': 'rsa',
     'keyid': keyid,
     'keyval': {'public': '-----BEGIN PUBLIC KEY----- ...',
                'private': ''}}

    The public portion of the RSA key is a string in PEM format.

    >>> rsa_key = generate_rsa_key()
    >>> public = rsa_key['keyval']['public']
    >>> rsa_key['keyval']['private'] = ''
    >>> rsa_key2 = import_rsakey_from_public_pem(public)
    >>> securesystemslib.formats.RSAKEY_SCHEMA.matches(rsa_key)
    True
    >>> securesystemslib.formats.RSAKEY_SCHEMA.matches(rsa_key2)
    True

  <Arguments>
    pem:
      A string in PEM format (it should contain a public RSA key).

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'pem' is improperly formatted.

  <Side Effects>
    Only the public portion of the PEM is extracted.  Leading or trailing
    whitespace is not included in the PEM string stored in the rsakey object
    returned.

  <Returns>
    A dictionary containing the RSA keys and other identifying information.
    Conforms to 'securesystemslib.formats.RSAKEY_SCHEMA'.
  """

  # Does 'pem' have the correct format?
  # This check will ensure arguments has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.PEMRSA_SCHEMA.check_match(pem)

  # Does 'scheme' have the correct format?
  formats.RSA_SCHEME_SCHEMA.check_match(scheme)

  # Ensure the PEM string has a public header and footer.  Although a simple
  # validation of 'pem' is performed here, a fully valid PEM string is needed
  # later to successfully verify signatures.  Performing stricter validation of
  # PEMs are left to the external libraries that use 'pem'.

  if is_pem_public(pem):
    public_pem = extract_pem(pem, private_pem=False)

  else:
    raise exceptions.FormatError('Invalid public pem: ' + repr(pem))

  # Begin building the RSA key dictionary.
  rsakey_dict = {}
  keytype = 'rsa'

  # Generate the keyid of the RSA key.  'key_value' corresponds to the
  # 'keyval' entry of the 'RSAKEY_SCHEMA' dictionary.  The private key
  # information is not included in the generation of the 'keyid' identifier.
  # Convert any '\r\n' (e.g., Windows) newline characters to '\n' so that a
  # consistent keyid is generated.
  key_value = {'public': public_pem.replace('\r\n', '\n'),
               'private': ''}
  keyid = _get_keyid(keytype, scheme, key_value)

  rsakey_dict['keytype'] = keytype
  rsakey_dict['scheme'] = scheme
  rsakey_dict['keyid'] = keyid
  rsakey_dict['keyval'] = key_value

  # Add "keyid_hash_algorithms" so that equal RSA keys with different keyids
  # can be associated using supported keyid_hash_algorithms.
  rsakey_dict['keyid_hash_algorithms'] = settings.HASH_ALGORITHMS

  return rsakey_dict





def import_rsakey_from_pem(pem, scheme='rsassa-pss-sha256'):
  """
  <Purpose>
    Import either a public or private PEM.  In contrast to the other explicit
    import functions (import_rsakey_from_public_pem and
    import_rsakey_from_private_pem), this function is useful for when it is not
    known whether 'pem' is private or public.

  <Arguments>
    pem:
      A string in PEM format.

    scheme:
      The signature scheme used by the imported key.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'pem' is improperly formatted.

  <Side Effects>
    None.

  <Returns>
    A dictionary containing the RSA keys and other identifying information.
    Conforms to 'securesystemslib.formats.RSAKEY_SCHEMA'.
  """

  # Does 'pem' have the correct format?
  # This check will ensure arguments has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.PEMRSA_SCHEMA.check_match(pem)

  # Is 'scheme' properly formatted?
  formats.RSA_SCHEME_SCHEMA.check_match(scheme)

  public_pem = ''

  # Ensure the PEM string has a public or private header and footer.  Although
  # a simple validation of 'pem' is performed here, a fully valid PEM string is
  # needed later to successfully verify signatures.  Performing stricter
  # validation of PEMs are left to the external libraries that use 'pem'.
  if is_pem_public(pem):
    public_pem = extract_pem(pem, private_pem=False)

  elif is_pem_private(pem):
    # Return an rsakey object (RSAKEY_SCHEMA) with the private key included.
    return import_rsakey_from_private_pem(pem, scheme, password=None)

  else:
    raise exceptions.FormatError('PEM contains neither a'
      ' public nor private key: ' + repr(pem))

  # Begin building the RSA key dictionary.
  rsakey_dict = {}
  keytype = 'rsa'

  # Generate the keyid of the RSA key.  'key_value' corresponds to the 'keyval'
  # entry of the 'RSAKEY_SCHEMA' dictionary.  The private key information is
  # not included in the generation of the 'keyid' identifier.  If a PEM is
  # found to contain a private key, the generated rsakey object should be
  # returned above.  The following key object is for the case of a PEM with
  # only a public key.  Convert any '\r\n' (e.g., Windows) newline characters
  # to '\n' so that a consistent keyid is generated.
  key_value = {'public': public_pem.replace('\r\n', '\n'),
               'private': ''}
  keyid = _get_keyid(keytype, scheme, key_value)

  rsakey_dict['keytype'] = keytype
  rsakey_dict['scheme'] = scheme
  rsakey_dict['keyid'] = keyid
  rsakey_dict['keyval'] = key_value

  # Add "keyid_hash_algorithms" so that equal RSA keys with
  # different keyids can be associated using supported keyid_hash_algorithms.
  rsakey_dict['keyid_hash_algorithms'] = settings.HASH_ALGORITHMS

  return rsakey_dict




def extract_pem(pem, private_pem=False):
  """
  <Purpose>
    Extract only the portion of the pem that includes the header and footer,
    with any leading and trailing characters removed.  The string returned has
    the following form:

    '-----BEGIN PUBLIC KEY----- ... -----END PUBLIC KEY-----'

    or

    '-----BEGIN RSA PRIVATE KEY----- ... -----END RSA PRIVATE KEY-----'

    Note: This function assumes "pem" is a valid pem in the following format:
    pem header + key material + key footer.  Crypto libraries (e.g., pyca's
    cryptography) that parse the pem returned by this function are expected to
    fully validate the pem.

  <Arguments>
    pem:
      A string in PEM format.

    private_pem:
      Boolean that indicates whether 'pem' is a private PEM.  Private PEMs
      are not shown in exception messages.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'pem' is improperly formatted.

  <Side Effects>
    Only the public and private portion of the PEM is extracted.  Leading or
    trailing whitespace is not included in the returned PEM string.

  <Returns>
    A PEM string (excluding leading and trailing newline characters).
    That is: pem header + key material + pem footer.

  """

  if private_pem:
    pem_header = '-----BEGIN RSA PRIVATE KEY-----'
    pem_footer = '-----END RSA PRIVATE KEY-----'

  else:
    pem_header = '-----BEGIN PUBLIC KEY-----'
    pem_footer = '-----END PUBLIC KEY-----'

  header_start = 0
  footer_start = 0

  # Raise error message if the expected header or footer is not found in 'pem'.
  try:
    header_start = pem.index(pem_header)

  except ValueError:
    # Be careful not to print private key material in exception message.
    if not private_pem:
      raise exceptions.FormatError('Required PEM'
        ' header ' + repr(pem_header) + '\n not found in PEM'
        ' string: ' + repr(pem))

    else:
      raise exceptions.FormatError('Required PEM'
        ' header ' + repr(pem_header) + '\n not found in private PEM string.')

  try:
    # Search for 'pem_footer' after the PEM header.
    footer_start = pem.index(pem_footer, header_start + len(pem_header))

  except ValueError:
    # Be careful not to print private key material in exception message.
    if not private_pem:
      raise exceptions.FormatError('Required PEM'
        ' footer ' + repr(pem_footer) + '\n not found in PEM'
        ' string ' + repr(pem))

    else:
      raise exceptions.FormatError('Required PEM'
        ' footer ' + repr(pem_footer) + '\n not found in private PEM string.')

  # Extract only the public portion of 'pem'.  Leading or trailing whitespace
  # is excluded.
  pem = pem[header_start:footer_start + len(pem_footer)]

  return pem





def encrypt_key(key_object, password):
  """
  <Purpose>
    Return a string containing 'key_object' in encrypted form. Encrypted
    strings may be safely saved to a file.  The corresponding decrypt_key()
    function can be applied to the encrypted string to restore the original key
    object.  'key_object' is a key (e.g., RSAKEY_SCHEMA, ED25519KEY_SCHEMA).
    This function relies on the rsa_keys.py module to perform the
    actual encryption.

    Encrypted keys use AES-256-CTR-Mode, and passwords are strengthened with
    PBKDF2-HMAC-SHA256 (100K iterations by default, but may be overriden in
    'securesystemslib.settings.PBKDF2_ITERATIONS' by the user).

    http://en.wikipedia.org/wiki/Advanced_Encryption_Standard
    http://en.wikipedia.org/wiki/CTR_mode#Counter_.28CTR.29
    https://en.wikipedia.org/wiki/PBKDF2

    >>> ed25519_key = generate_ed25519_key()
    >>> password = 'secret'
    >>> encrypted_key = encrypt_key(ed25519_key, password).encode('utf-8')
    >>> securesystemslib.formats.ENCRYPTEDKEY_SCHEMA.matches(encrypted_key)
    True

  <Arguments>
    key_object:
      A key (containing also the private key portion) of the form
      'securesystemslib.formats.ANYKEY_SCHEMA'

    password:
      The password, or passphrase, to encrypt the private part of the RSA
      key.  'password' is not used directly as the encryption key, a stronger
      encryption key is derived from it.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.CryptoError, if 'key_object' cannot be
    encrypted.

  <Side Effects>
    None.

  <Returns>
    An encrypted string of the form:
    'securesystemslib.formats.ENCRYPTEDKEY_SCHEMA'.
  """

  # Does 'key_object' have the correct format?
  # This check will ensure 'key_object' has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.ANYKEY_SCHEMA.check_match(key_object)

  # Does 'password' have the correct format?
  formats.PASSWORD_SCHEMA.check_match(password)

  # Encrypted string of 'key_object'.  The encrypted string may be safely saved
  # to a file and stored offline.
  encrypted_key = None

  # Generate an encrypted string of 'key_object' using AES-256-CTR-Mode, where
  # 'password' is strengthened with PBKDF2-HMAC-SHA256.
  encrypted_key = rsa_keys.encrypt_key(key_object, password)

  return encrypted_key





def decrypt_key(encrypted_key, passphrase):
  """
  <Purpose>
    Return a string containing 'encrypted_key' in non-encrypted form.  The
    decrypt_key() function can be applied to the encrypted string to restore
    the original key object, a key (e.g., RSAKEY_SCHEMA, ED25519KEY_SCHEMA).
    This function calls rsa_keys.py to perform the actual decryption.

    Encrypted keys use AES-256-CTR-Mode and passwords are strengthened with
    PBKDF2-HMAC-SHA256 (100K iterations be default, but may be overriden in
    'settings.py' by the user).

    http://en.wikipedia.org/wiki/Advanced_Encryption_Standard
    http://en.wikipedia.org/wiki/CTR_mode#Counter_.28CTR.29
    https://en.wikipedia.org/wiki/PBKDF2

    >>> ed25519_key = generate_ed25519_key()
    >>> password = 'secret'
    >>> encrypted_key = encrypt_key(ed25519_key, password)
    >>> decrypted_key = decrypt_key(encrypted_key.encode('utf-8'), password)
    >>> securesystemslib.formats.ANYKEY_SCHEMA.matches(decrypted_key)
    True
    >>> decrypted_key == ed25519_key
    True

  <Arguments>
    encrypted_key:
      An encrypted key (additional data is also included, such as salt, number
      of password iterations used for the derived encryption key, etc) of the
      form 'securesystemslib.formats.ENCRYPTEDKEY_SCHEMA'.  'encrypted_key'
      should have been generated with encrypt_key().

    password:
      The password, or passphrase, to decrypt 'encrypted_key'.  'password' is
      not used directly as the encryption key, a stronger encryption key is
      derived from it.  The supported general-purpose module takes care of
      re-deriving the encryption key.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.CryptoError, if 'encrypted_key' cannot be
    decrypted.

  <Side Effects>
    None.

  <Returns>
    A key object of the form: 'securesystemslib.formats.ANYKEY_SCHEMA' (e.g.,
    RSAKEY_SCHEMA, ED25519KEY_SCHEMA).
  """

  # Does 'encrypted_key' have the correct format?
  # This check ensures 'encrypted_key' has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.ENCRYPTEDKEY_SCHEMA.check_match(encrypted_key)

  # Does 'passphrase' have the correct format?
  formats.PASSWORD_SCHEMA.check_match(passphrase)

  # Store and return the decrypted key object.
  key_object = None

  # Decrypt 'encrypted_key' so that the original key object is restored.
  # encrypt_key() generates an encrypted string of the key object using
  # AES-256-CTR-Mode, where 'password' is strengthened with PBKDF2-HMAC-SHA256.
  key_object = rsa_keys.decrypt_key(encrypted_key, passphrase)

  # The corresponding encrypt_key() encrypts and stores key objects in
  # non-metadata format (i.e., original format of key object argument to
  # encrypt_key()) prior to returning.

  return key_object





def create_rsa_encrypted_pem(private_key, passphrase):
  """
  <Purpose>
    Return a string in PEM format (TraditionalOpenSSL), where the private part
    of the RSA key is encrypted using the best available encryption for a given
    key's backend. This is a curated (by cryptography.io) encryption choice and
    the algorithm may change over time.

    c.f. cryptography.io/en/latest/hazmat/primitives/asymmetric/serialization/
        #cryptography.hazmat.primitives.serialization.BestAvailableEncryption

  >>> rsa_key = generate_rsa_key()
  >>> private = rsa_key['keyval']['private']
  >>> passphrase = 'secret'
  >>> encrypted_pem = create_rsa_encrypted_pem(private, passphrase)
  >>> securesystemslib.formats.PEMRSA_SCHEMA.matches(encrypted_pem)
  True

  <Arguments>
    private_key:
      The private key string in PEM format.

    passphrase:
      The passphrase, or password, to encrypt the private part of the RSA key.
      'passphrase' is not used directly as the encryption key, a stronger
      encryption key is derived from it.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.CryptoError, if an RSA key in encrypted PEM
    format cannot be created.

    TypeError, 'private_key' is unset.

  <Side Effects>
    None.

  <Returns>
    A string in PEM format, where the private RSA key is encrypted.
    Conforms to 'securesystemslib.formats.PEMRSA_SCHEMA'.
  """

  # Does 'private_key' have the correct format?
  # This check will ensure 'private_key' has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.PEMRSA_SCHEMA.check_match(private_key)

  # Does 'passphrase' have the correct format?
  formats.PASSWORD_SCHEMA.check_match(passphrase)

  encrypted_pem = None

  # Generate the public and private RSA keys. A 2048-bit minimum is enforced by
  # create_rsa_encrypted_pem() via a
  # securesystemslib.formats.RSAKEYBITS_SCHEMA.check_match().
  encrypted_pem = rsa_keys.create_rsa_encrypted_pem(private_key, passphrase)

  return encrypted_pem




def is_pem_public(pem):
  """
  <Purpose>
    Checks if a passed PEM formatted string is a PUBLIC key, by looking for the
    following pattern:

    '-----BEGIN PUBLIC KEY----- ... -----END PUBLIC KEY-----'

    >>> rsa_key = generate_rsa_key()
    >>> public = rsa_key['keyval']['public']
    >>> private = rsa_key['keyval']['private']
    >>> is_pem_public(public)
    True
    >>> is_pem_public(private)
    False

  <Arguments>
    pem:
      A string in PEM format.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'pem' is improperly formatted.

  <Side Effects>
    None

  <Returns>
    True if 'pem' is public and false otherwise.
  """

  # Do the arguments have the correct format?
  # This check will ensure arguments have the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.PEMRSA_SCHEMA.check_match(pem)

  pem_header = '-----BEGIN PUBLIC KEY-----'
  pem_footer = '-----END PUBLIC KEY-----'

  try:
    header_start = pem.index(pem_header)
    pem.index(pem_footer, header_start + len(pem_header))

  except ValueError:
    return False

  return True




def is_pem_private(pem, keytype='rsa'):
  """
  <Purpose>
    Checks if a passed PEM formatted string is a PRIVATE key, by looking for
    the following patterns:

    '-----BEGIN RSA PRIVATE KEY----- ... -----END RSA PRIVATE KEY-----'
    '-----BEGIN EC PRIVATE KEY----- ... -----END EC PRIVATE KEY-----'

    >>> rsa_key = generate_rsa_key()
    >>> private = rsa_key['keyval']['private']
    >>> public = rsa_key['keyval']['public']
    >>> is_pem_private(private)
    True
    >>> is_pem_private(public)
    False

  <Arguments>
    pem:
      A string in PEM format.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if any of the arguments are
    improperly formatted.

  <Side Effects>
    None

  <Returns>
    True if 'pem' is private and false otherwise.
  """

  # Do the arguments have the correct format?
  # This check will ensure arguments have the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.PEMRSA_SCHEMA.check_match(pem)
  formats.NAME_SCHEMA.check_match(keytype)

  if keytype == 'rsa':
    pem_header = '-----BEGIN RSA PRIVATE KEY-----'
    pem_footer = '-----END RSA PRIVATE KEY-----'

  elif keytype == 'ec':
    pem_header = '-----BEGIN EC PRIVATE KEY-----'
    pem_footer = '-----END EC PRIVATE KEY-----'

  else:
    raise exceptions.FormatError('Unsupported key'
      ' type: ' + repr(keytype) + '.  Supported keytypes: ["rsa", "ec"]')

  try:
    header_start = pem.index(pem_header)
    pem.index(pem_footer, header_start + len(pem_header))

  except ValueError:
    return False

  return True





def import_ed25519key_from_private_json(json_str, password=None):
  if password is not None:
    # This check will not fail, because a mal-formatted passed password fails
    # above and an entered password will always be a string (see get_password)
    # However, we include it in case PASSWORD_SCHEMA or get_password changes.
    formats.PASSWORD_SCHEMA.check_match(password)

    # Decrypt the loaded key file, calling the 'cryptography' library to
    # generate the derived encryption key from 'password'.  Raise
    # 'securesystemslib.exceptions.CryptoError' if the decryption fails.
    key_object = decrypt_key(json_str.decode('utf-8'), password)

  else:
    logger.debug('No password was given. Attempting to import an'
        ' unencrypted file.')
    try:
      key_object = util.load_json_string(json_str.decode('utf-8'))
    # If the JSON could not be decoded, it is very likely, but not necessarily,
    # due to a non-empty password.
    except exceptions.Error:
      raise exceptions.CryptoError('Malformed Ed25519 key JSON, '
                                   'possibly due to encryption, '
                                   'but no password provided?')

  # Raise an exception if an unexpected key type is imported.
  if key_object['keytype'] != 'ed25519':
    message = 'Invalid key type loaded: ' + repr(key_object['keytype'])
    raise exceptions.FormatError(message)

  # Add "keyid_hash_algorithms" so that equal ed25519 keys with
  # different keyids can be associated using supported keyid_hash_algorithms.
  key_object['keyid_hash_algorithms'] = settings.HASH_ALGORITHMS

  return key_object





def import_ecdsakey_from_private_pem(pem, scheme='ecdsa-sha2-nistp256', password=None):
  """
  <Purpose>
    Import the private ECDSA key stored in 'pem', and generate its public key
    (which will also be included in the returned ECDSA key object).  In addition,
    a keyid identifier for the ECDSA key is generated.  The object returned
    conforms to:

    {'keytype': 'ecdsa',
     'scheme': 'ecdsa-sha2-nistp256',
     'keyid': keyid,
     'keyval': {'public': '-----BEGIN PUBLIC KEY----- ... -----END PUBLIC KEY-----',
                'private': '-----BEGIN EC PRIVATE KEY----- ... -----END EC PRIVATE KEY-----'}}

    The private key is a string in PEM format.

    >>> ecdsa_key = generate_ecdsa_key()
    >>> private_pem = ecdsa_key['keyval']['private']
    >>> ecdsa_key = import_ecdsakey_from_private_pem(private_pem)
    >>> securesystemslib.formats.ECDSAKEY_SCHEMA.matches(ecdsa_key)
    True

  <Arguments>
    pem:
      A string in PEM format.  The private key is extracted and returned in
      an ecdsakey object.

    scheme:
      The signature scheme used by the imported key.

    password: (optional)
      The password, or passphrase, to decrypt the private part of the ECDSA
      key if it is encrypted.  'password' is not used directly as the encryption
      key, a stronger encryption key is derived from it.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the arguments are improperly
    formatted.

    securesystemslib.exceptions.UnsupportedAlgorithmError, if 'pem' specifies
    an unsupported key type.

  <Side Effects>
    None.

  <Returns>
    A dictionary containing the ECDSA keys and other identifying information.
    Conforms to 'securesystemslib.formats.ECDSAKEY_SCHEMA'.
  """

  # Does 'pem' have the correct format?
  # This check will ensure 'pem' conforms to
  # 'securesystemslib.formats.ECDSARSA_SCHEMA'.
  formats.PEMECDSA_SCHEMA.check_match(pem)

  # Is 'scheme' properly formatted?
  formats.ECDSA_SCHEME_SCHEMA.check_match(scheme)

  if password is not None:
    formats.PASSWORD_SCHEMA.check_match(password)

  else:
    logger.debug('The password/passphrase is unset.  The PEM is expected'
      ' to be unencrypted.')

  # Begin building the ECDSA key dictionary.
  ecdsakey_dict = {}
  keytype = 'ecdsa'
  public = None
  private = None

  public, private = \
      ecdsa_keys.create_ecdsa_public_and_private_from_pem(pem, password)

  # Generate the keyid of the ECDSA key.  'key_value' corresponds to the
  # 'keyval' entry of the 'ECDSAKEY_SCHEMA' dictionary.  The private key
  # information is not included in the generation of the 'keyid' identifier.
  # Convert any '\r\n' (e.g., Windows) newline characters to '\n' so that a
  # consistent keyid is generated.
  key_value = {'public': public.replace('\r\n', '\n'),
               'private': ''}
  keyid = _get_keyid(keytype, scheme, key_value)

  # Build the 'ecdsakey_dict' dictionary.  Update 'key_value' with the ECDSA
  # private key prior to adding 'key_value' to 'ecdsakey_dict'.
  key_value['private'] = private

  ecdsakey_dict['keytype'] = keytype
  ecdsakey_dict['scheme'] = scheme
  ecdsakey_dict['keyid'] = keyid
  ecdsakey_dict['keyval'] = key_value

  # Add "keyid_hash_algorithms" so equal ECDSA keys with
  # different keyids can be associated using supported keyid_hash_algorithms
  ecdsakey_dict['keyid_hash_algorithms'] = settings.HASH_ALGORITHMS

  return ecdsakey_dict





def import_ecdsakey_from_public_pem(pem, scheme='ecdsa-sha2-nistp256'):
  """
  <Purpose>
    Generate an ECDSA key object from 'pem'.  In addition, a keyid identifier
    for the ECDSA key is generated.  The object returned conforms to
    'securesystemslib.formats.ECDSAKEY_SCHEMA' and has the form:

    {'keytype': 'ecdsa',
     'scheme': 'ecdsa-sha2-nistp256',
     'keyid': keyid,
     'keyval': {'public': '-----BEGIN PUBLIC KEY----- ...',
                'private': ''}}

    The public portion of the ECDSA key is a string in PEM format.

    >>> ecdsa_key = generate_ecdsa_key()
    >>> public = ecdsa_key['keyval']['public']
    >>> ecdsa_key['keyval']['private'] = ''
    >>> scheme = ecdsa_key['scheme']
    >>> ecdsa_key2 = import_ecdsakey_from_public_pem(public, scheme)
    >>> securesystemslib.formats.ECDSAKEY_SCHEMA.matches(ecdsa_key)
    True
    >>> securesystemslib.formats.ECDSAKEY_SCHEMA.matches(ecdsa_key2)
    True

  <Arguments>
    pem:
      A string in PEM format (it should contain a public ECDSA key).

    scheme:
      The signature scheme used by the imported key.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'pem' is improperly formatted.

  <Side Effects>
    Only the public portion of the PEM is extracted.  Leading or trailing
    whitespace is not included in the PEM string stored in the rsakey object
    returned.

  <Returns>
    A dictionary containing the ECDSA keys and other identifying information.
    Conforms to 'securesystemslib.formats.ECDSAKEY_SCHEMA'.
  """

  # Does 'pem' have the correct format?
  # This check will ensure arguments has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.PEMECDSA_SCHEMA.check_match(pem)

  # Is 'scheme' properly formatted?
  formats.ECDSA_SCHEME_SCHEMA.check_match(scheme)

  # Ensure the PEM string has a public header and footer.  Although a simple
  # validation of 'pem' is performed here, a fully valid PEM string is needed
  # later to successfully verify signatures.  Performing stricter validation of
  # PEMs are left to the external libraries that use 'pem'.

  if is_pem_public(pem):
    public_pem = extract_pem(pem, private_pem=False)

  else:
    raise exceptions.FormatError('Invalid public'
        ' pem: ' + repr(pem))

  # Begin building the ECDSA key dictionary.
  ecdsakey_dict = {}
  keytype = 'ecdsa'

  # Generate the keyid of the ECDSA key.  'key_value' corresponds to the
  # 'keyval' entry of the 'ECDSAKEY_SCHEMA' dictionary.  The private key
  # information is not included in the generation of the 'keyid' identifier.
  # Convert any '\r\n' (e.g., Windows) newline characters to '\n' so that a
  # consistent keyid is generated.
  key_value = {'public': public_pem.replace('\r\n', '\n'),
               'private': ''}
  keyid = _get_keyid(keytype, scheme, key_value)

  ecdsakey_dict['keytype'] = keytype
  ecdsakey_dict['scheme'] = scheme
  ecdsakey_dict['keyid'] = keyid
  ecdsakey_dict['keyval'] = key_value

  # Add "keyid_hash_algorithms" so that equal ECDSA keys with different keyids
  # can be associated using supported keyid_hash_algorithms.
  ecdsakey_dict['keyid_hash_algorithms'] = settings.HASH_ALGORITHMS

  return ecdsakey_dict





def import_ecdsakey_from_pem(pem, scheme='ecdsa-sha2-nistp256'):
  """
  <Purpose>
    Import either a public or private ECDSA PEM.  In contrast to the other
    explicit import functions (import_ecdsakey_from_public_pem and
    import_ecdsakey_from_private_pem), this function is useful for when it is
    not known whether 'pem' is private or public.

  <Arguments>
    pem:
      A string in PEM format.

    scheme:
      The signature scheme used by the imported key.
  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'pem' is improperly formatted.

  <Side Effects>
    None.

  <Returns>
    A dictionary containing the ECDSA keys and other identifying information.
    Conforms to 'securesystemslib.formats.ECDSAKEY_SCHEMA'.
  """

  # Does 'pem' have the correct format?
  # This check will ensure arguments has the appropriate number
  # of objects and object types, and that all dict keys are properly named.
  # Raise 'securesystemslib.exceptions.FormatError' if the check fails.
  formats.PEMECDSA_SCHEMA.check_match(pem)

  # Is 'scheme' properly formatted?
  formats.ECDSA_SCHEME_SCHEMA.check_match(scheme)

  public_pem = ''

  # Ensure the PEM string has a public or private header and footer.  Although
  # a simple validation of 'pem' is performed here, a fully valid PEM string is
  # needed later to successfully verify signatures.  Performing stricter
  # validation of PEMs are left to the external libraries that use 'pem'.
  if is_pem_public(pem):
    public_pem = extract_pem(pem, private_pem=False)

  elif is_pem_private(pem, 'ec'):
    # Return an ecdsakey object (ECDSAKEY_SCHEMA) with the private key included.
    return import_ecdsakey_from_private_pem(pem, password=None)

  else:
    raise exceptions.FormatError('PEM contains neither a public'
      ' nor private key: ' + repr(pem))

  # Begin building the ECDSA key dictionary.
  ecdsakey_dict = {}
  keytype = 'ecdsa'

  # Generate the keyid of the ECDSA key.  'key_value' corresponds to the
  # 'keyval' entry of the 'ECDSAKEY_SCHEMA' dictionary.  The private key
  # information is not included in the generation of the 'keyid' identifier.
  # If a PEM is found to contain a private key, the generated rsakey object
  # should be returned above.  The following key object is for the case of a
  # PEM with only a public key.  Convert any '\r\n' (e.g., Windows) newline
  # characters to '\n' so that a consistent keyid is generated.
  key_value = {'public': public_pem.replace('\r\n', '\n'),
               'private': ''}
  keyid = _get_keyid(keytype, scheme, key_value)

  ecdsakey_dict['keytype'] = keytype
  ecdsakey_dict['scheme'] = scheme
  ecdsakey_dict['keyid'] = keyid
  ecdsakey_dict['keyval'] = key_value

  return ecdsakey_dict



if __name__ == '__main__':
  # The interactive sessions of the documentation strings can
  # be tested by running 'keys.py' as a standalone module:
  # $ python keys.py
  import doctest
  doctest.testmod()



import logging

# Configure a basic 'securesystemslib' top-level logger with a StreamHandler
# (print to console) and the WARNING log level (print messages of type
# warning, error or critical). This is similar to what 'logging.basicConfig'
# would do with the root logger. All 'securesystemslib.*' loggers default to
# this top-level logger and thus may be configured (e.g. formatted, silenced,
# etc.) with it. It can be accessed via logging.getLogger('securesystemslib').
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
logger.addHandler(logging.StreamHandler())


# Global constants
# TODO: Replace hard-coded key types with these constants (and add more)
KEY_TYPE_RSA = "rsa"
KEY_TYPE_ED25519 = "ed25519"
KEY_TYPE_ECDSA = "ecdsa"



"""
<Program Name>
  test_storage.py

<Author>
  Joshua Lock <jlock@vmware.com>

<Started>
  April 17, 2020

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Unit test for 'storage.py'
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os
import tempfile
import shutil
import unittest

import securesystemslib.exceptions
import securesystemslib.storage

import six


class TestStorage(unittest.TestCase):
  def setUp(self):
    self.storage_backend = securesystemslib.storage.FilesystemBackend()
    self.temp_dir = tempfile.mkdtemp(dir=os.getcwd())
    self.filepath = os.path.join(self.temp_dir, 'testfile')
    with open(self.filepath, 'wb') as test:
        test.write(b'testing')
    self.fileobj = open(self.filepath, 'rb')


  def tearDown(self):
    self.fileobj.close()
    shutil.rmtree(self.temp_dir)


  def test_exceptions(self):
    try:
      with self.storage_backend.get('/none/existent/path') as file_object:
        file_object.read()
    except Exception as exc:
      self.assertIsInstance(exc, securesystemslib.exceptions.StorageError)

    self.assertRaises(securesystemslib.exceptions.StorageError,
        self.storage_backend.put, self.fileobj, '/none/existent/path')

    self.assertRaises(securesystemslib.exceptions.StorageError,
        self.storage_backend.getsize, '/none/existent/path')

    self.assertRaises(securesystemslib.exceptions.StorageError,
        self.storage_backend.create_folder, '/none/existent/path')

    self.assertRaises(securesystemslib.exceptions.StorageError,
        self.storage_backend.create_folder, '')

    self.assertRaises(securesystemslib.exceptions.StorageError,
        self.storage_backend.list_folder, '/none/existent/path')


  def test_files(self):
    with self.storage_backend.get(self.filepath) as get_fileobj:
      self.assertEqual(get_fileobj.read(), self.fileobj.read())

    self.assertEqual(self.storage_backend.getsize(self.filepath),
        os.path.getsize(self.filepath))

    put_path = os.path.join(self.temp_dir, 'put')
    with self.storage_backend.get(self.filepath) as get_fileobj:
      self.storage_backend.put(get_fileobj, put_path)
      self.fileobj.seek(0)
      with open(put_path, 'rb') as put_file:
        self.assertEqual(put_file.read(), self.fileobj.read())

    self.assertTrue(os.path.exists(put_path))
    self.storage_backend.remove(put_path)
    self.assertFalse(os.path.exists(put_path))


  def test_folders(self):
    leaves = ['test1', 'test2', 'test3']
    folder = os.path.join(self.temp_dir, 'test_dir')
    self.storage_backend.create_folder(folder)
    for leaf in leaves:
      with open(os.path.join(folder, leaf), 'wb') as fi:
        fi.write(leaf.encode('utf-8'))
    found_leaves = self.storage_backend.list_folder(folder)
    self.assertListEqual(leaves, sorted(found_leaves))


  def test_singleton(self):
    # There should only ever be a single instance of FilesystemBackend.
    # An object's id is unique and constant for the object during its
    # lifetime. Therefore create more than one instance of FilesystemBackend
    # and compare their id's
    fb1 = securesystemslib.storage.FilesystemBackend()
    fb2 = securesystemslib.storage.FilesystemBackend()
    self.assertEqual(id(fb1), id(fb2))
    self.assertEqual(id(self.storage_backend), id(fb1))
    self.assertEqual(id(fb2), id(self.storage_backend))



"""
<Program Name>
  exceptions.py

<Author>
  Santiago Torres-Arias <santiago@nyu.edu>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  Dec 8, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Define Exceptions used in the gpg package. Following the practice from
  securesystemslib the names chosen for exception classes should end in
  'Error' (except where there is a good reason not to).

"""
import datetime


class PacketParsingError(Exception):
  pass

class KeyNotFoundError(Exception):
  pass

class PacketVersionNotSupportedError(Exception):
  pass

class SignatureAlgorithmNotSupportedError(Exception):
  pass

class CommandError(Exception):
  pass

class KeyExpirationError(Exception):
  def __init__(self, key):
    super(KeyExpirationError, self).__init__()
    self.key = key

  def __str__(self):
    creation_time = datetime.datetime.utcfromtimestamp(
        self.key["creation_time"])
    expiration_time = datetime.datetime.utcfromtimestamp(
        self.key["creation_time"] + self.key["validity_period"])
    validity_period = expiration_time - creation_time

    return ("GPG key '{}' created on '{:%Y-%m-%d %H:%M} UTC' with validity "
        "period '{}' expired on '{:%Y-%m-%d %H:%M} UTC'.".format(
        self.key["keyid"], creation_time, validity_period, expiration_time))



"""
<Module Name>
  util.py

<Author>
  Santiago Torres-Arias <santiago@nyu.edu>

<Started>
  Nov 15, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  general-purpose utilities for binary data handling and pgp data parsing
"""
import struct
import binascii
import re
import logging

from distutils.version import StrictVersion # pylint: disable=no-name-in-module,import-error

CRYPTO = True
NO_CRYPTO_MSG = 'gpg.utils requires the cryptography library'
try:
  from cryptography.hazmat import backends
  from cryptography.hazmat.primitives import hashes as hashing
except ImportError:
  CRYPTO = False

from securesystemslib import exceptions
from securesystemslib import process
from securesystemslib.gpg import constants
from securesystemslib.gpg.exceptions import PacketParsingError

log = logging.getLogger(__name__)


def get_mpi_length(data):
  """
  <Purpose>
    parses an MPI (Multi-Precision Integer) buffer and returns the appropriate
    length. This is mostly done to perform bitwise to byte-wise conversion.

    See RFC4880 section 3.2. Multiprecision Integers for details.

  <Arguments>
    data: The MPI data

  <Exceptions>
    None

  <Side Effects>
    None

  <Returns>
    The length of the MPI contained at the beginning of this data buffer.
  """
  bitlength = int(struct.unpack(">H", data)[0])
  # Notice the /8 at the end, this length is the bitlength, not the length of
  # the data in bytes (as len reports it)
  return int((bitlength - 1)/8) + 1


def hash_object(headers, algorithm, content):
  """
  <Purpose>
    Hash data prior to signature verification in conformance of the RFC4880
    openPGP standard.

  <Arguments>
    headers: the additional OpenPGP headers as populated from
    gpg_generate_signature

    algorithm: The hash algorithm object defined by the cryptography.io hashes
    module

    content: the signed content

  <Exceptions>
    securesystemslib.exceptions.UnsupportedLibraryError if:
      the cryptography module is unavailable

  <Side Effects>
    None

  <Returns>
    The RFC4880-compliant hashed buffer
  """
  if not CRYPTO: # pragma: no cover
    raise exceptions.UnsupportedLibraryError(NO_CRYPTO_MSG)

  # As per RFC4880 Section 5.2.4., we need to hash the content,
  # signature headers and add a very opinionated trailing header
  hasher = hashing.Hash(algorithm, backend=backends.default_backend())
  hasher.update(content)
  hasher.update(headers)
  hasher.update(b'\x04\xff')
  hasher.update(struct.pack(">I", len(headers)))

  return hasher.finalize()


def parse_packet_header(data, expected_type=None):
  """
  <Purpose>
    Parse out packet type and header and body lengths from an RFC4880 packet.

  <Arguments>
    data:
            An RFC4880 packet as described in section 4.2 of the rfc.

    expected_type: (optional)
            Used to error out if the packet does not have the expected
            type. See securesystemslib.gpg.constants.PACKET_TYPE_* for
            available types.

  <Exceptions>
    securesystemslib.gpg.exceptions.PacketParsingError
            If the new format packet length encodes a partial body length
            If the old format packet length encodes an indeterminate length
            If header or body length could not be determined
            If the expected_type was passed and does not match the packet type

    IndexError
            If the passed data is incomplete

  <Side Effects>
    None.

  <Returns>
    A tuple of packet type, header length, body length and packet length.
    (see  RFC4880 4.3. for the list of available packet types)

  """
  data = bytearray(data)
  header_len = None
  body_len = None

  # If Bit 6 of 1st octet is set we parse a New Format Packet Length, and
  # an Old Format Packet Lengths otherwise
  if data[0] & 0b01000000:
    # In new format packet lengths the packet type is encoded in Bits 5-0 of
    # the 1st octet of the packet
    packet_type = data[0] & 0b00111111

    # The rest of the packet header is the body length header, which may
    # consist of one, two or five octets. To disambiguate the RFC, the first
    # octet of the body length header is the second octet of the packet.
    if data[1] < 192:
      header_len = 2
      body_len = data[1]

    elif data[1] >= 192 and data[1] <= 223:
      header_len = 3
      body_len = (data[1] - 192 << 8) + data[2] + 192

    elif data[1] >= 224 and data[1] < 255:
      raise PacketParsingError("New length "
          "format packets of partial body lengths are not supported")

    elif data[1] == 255:
      header_len = 6
      body_len = data[2] << 24 | data[3] << 16 | data[4] << 8 | data[5]

    else: # pragma: no cover
      # Unreachable: octet must be between 0 and 255
      raise PacketParsingError("Invalid new length")

  else:
    # In old format packet lengths the packet type is encoded in Bits 5-2 of
    # the 1st octet and the length type in Bits 1-0
    packet_type = (data[0] & 0b00111100) >> 2
    length_type = data[0] & 0b00000011

    # The body length is encoded using one, two, or four octets, starting
    # with the second octet of the packet
    if length_type == 0:
      body_len = data[1]
      header_len = 2

    elif length_type == 1:
      header_len = 3
      body_len = struct.unpack(">H", data[1:header_len])[0]

    elif length_type == 2:
      header_len = 5
      body_len = struct.unpack(">I", data[1:header_len])[0]

    elif length_type == 3:
      raise PacketParsingError("Old length "
          "format packets of indeterminate length are not supported")

    else: # pragma: no cover (unreachable)
      # Unreachable: bits 1-0 must be one of 0 to 3
      raise PacketParsingError("Invalid old length")

  if header_len is None or body_len is None: # pragma: no cover
    # Unreachable: One of above must have assigned lengths or raised error
    raise PacketParsingError("Could not determine packet length")

  if expected_type is not None and packet_type != expected_type:
    raise PacketParsingError("Expected packet "
        "{}, but got {} instead!".format(expected_type, packet_type))

  return packet_type, header_len, body_len, header_len + body_len


def compute_keyid(pubkey_packet_data):
  """
  <Purpose>
    compute a keyid from an RFC4880 public-key buffer

  <Arguments>
    pubkey_packet_data: the public-key packet buffer

  <Exceptions>
    securesystemslib.exceptions.UnsupportedLibraryError if:
      the cryptography module is unavailable

  <Side Effects>
    None

  <Returns>
    The RFC4880-compliant hashed buffer
  """
  if not CRYPTO: # pragma: no cover
    raise exceptions.UnsupportedLibraryError(NO_CRYPTO_MSG)

  hasher = hashing.Hash(hashing.SHA1(), backend=backends.default_backend())
  hasher.update(b'\x99')
  hasher.update(struct.pack(">H", len(pubkey_packet_data)))
  hasher.update(bytes(pubkey_packet_data))
  return binascii.hexlify(hasher.finalize()).decode("ascii")


def parse_subpacket_header(data):
  """ Parse out subpacket header as per RFC4880 5.2.3.1. Signature Subpacket
  Specification. """
  # NOTE: Although the RFC does not state it explicitly, the length encoded
  # in the header must be greater equal 1, as it includes the mandatory
  # subpacket type octet.
  # Hence, passed bytearrays like [0] or [255, 0, 0, 0, 0], which encode a
  # subpacket length 0  are invalid.
  # The caller has to deal with the resulting IndexError.
  if data[0] < 192:
    length_len = 1
    length = data[0]

  elif data[0] >= 192 and data[0] < 255:
    length_len = 2
    length = ((data[0] - 192 << 8) + (data[1] + 192))

  elif data[0] == 255:
    length_len = 5
    length = struct.unpack(">I", data[1:length_len])[0]

  else: # pragma: no cover (unreachable)
    raise PacketParsingError("Invalid subpacket header")

  return data[length_len], length_len + 1, length - 1, length_len + length

def parse_subpackets(data):
  """
  <Purpose>
    parse the subpackets fields

  <Arguments>
    data: the unparsed subpacketoctets

  <Exceptions>
    IndexErrorif the subpackets octets are incomplete or malformed

  <Side Effects>
    None

  <Returns>
    A list of tuples with like:
        [ (packet_type, data),
          (packet_type, data),
          ...
        ]
  """
  parsed_subpackets = []
  position = 0

  while position < len(data):
    subpacket_type, header_len, _, subpacket_len = \
        parse_subpacket_header(data[position:])

    payload = data[position+header_len:position+subpacket_len]
    parsed_subpackets.append((subpacket_type, payload))

    position += subpacket_len

  return parsed_subpackets


def get_version():
  """
  <Purpose>
    Uses `gpg2 --version` to get the version info of the installed gpg2
    and extracts and returns the version number.

    The executed base command is defined in constants.GPG_VERSION_COMMAND.

  <Exceptions>
    securesystemslib.exceptions.UnsupportedLibraryError:
            If the gpg command is not available

  <Returns>
    Version number string, e.g. "2.1.22"

  """
  if not constants.HAVE_GPG: # pragma: no cover
    raise exceptions.UnsupportedLibraryError(constants.NO_GPG_MSG)

  command = constants.GPG_VERSION_COMMAND
  gpg_process = process.run(command, stdout=process.PIPE,
      stderr=process.PIPE, universal_newlines=True)

  full_version_info = gpg_process.stdout
  version_string = re.search(r'(\d\.\d\.\d+)', full_version_info).group(1)

  return version_string


def is_version_fully_supported():
  """
  <Purpose>
    Compares the version of installed gpg2 with the minimal fully supported
    gpg2 version (2.1.0).

  <Returns>
    True if the version returned by `get_version` is greater-equal
    constants.FULLY_SUPPORTED_MIN_VERSION, False otherwise.

  """

  installed_version = get_version()
  # Excluded so that coverage does not vary in different test environments
  if (StrictVersion(installed_version) >=
      StrictVersion(constants.FULLY_SUPPORTED_MIN_VERSION)): # pragma: no cover
    return True

  else: # pragma: no cover
    return False


def get_hashing_class(hash_algorithm_id):
  """
  <Purpose>
    Return a pyca/cryptography hashing class reference for the passed RFC4880
    hash algorithm ID.

  <Arguments>
    hash_algorithm_id:
            one of SHA1, SHA256, SHA512 (see securesystemslib.gpg.constants)

  <Exceptions>
    ValueError
            if the passed hash_algorithm_id is not supported.

  <Returns>
    A pyca/cryptography hashing class

  """
  supported_hashing_algorithms = [constants.SHA1, constants.SHA256,
      constants.SHA512]
  corresponding_hashing_classes = [hashing.SHA1, hashing.SHA256,
      hashing.SHA512]

  # Map supported hash algorithm ids to corresponding hashing classes
  hashing_class = dict(zip(supported_hashing_algorithms,
      corresponding_hashing_classes))

  try:
    return hashing_class[hash_algorithm_id]

  except KeyError:
    raise ValueError("Hash algorithm '{}' not supported, must be one of '{}' "
        "(see RFC4880 9.4. Hash Algorithms).".format(hash_algorithm_id,
        supported_hashing_algorithms))



#!/usr/bin/env python

"""
<Program Name>
  test_interface.py

<Author>
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  January 5, 2017.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Unit test for 'interface.py'.
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os
import time
import datetime
import tempfile
import json
import shutil
import stat
import sys
import unittest

from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.backends import default_backend

# Use external backport 'mock' on versions under 3.3
if sys.version_info >= (3, 3):
  import unittest.mock as mock

else:
  import mock

from securesystemslib import (
    KEY_TYPE_RSA,
    KEY_TYPE_ED25519,
    KEY_TYPE_ECDSA)

from securesystemslib.formats import (
    RSAKEY_SCHEMA,
    PUBLIC_KEY_SCHEMA,
    ANY_PUBKEY_DICT_SCHEMA,
    ED25519KEY_SCHEMA,
    ECDSAKEY_SCHEMA)

from securesystemslib.exceptions import Error, FormatError, CryptoError

from securesystemslib.interface import (
    _generate_and_write_rsa_keypair,
    generate_and_write_rsa_keypair,
    generate_and_write_rsa_keypair_with_prompt,
    generate_and_write_unencrypted_rsa_keypair,
    import_rsa_privatekey_from_file,
    import_rsa_publickey_from_file,
    _generate_and_write_ed25519_keypair,
    generate_and_write_ed25519_keypair,
    generate_and_write_ed25519_keypair_with_prompt,
    generate_and_write_unencrypted_ed25519_keypair,
    import_ed25519_publickey_from_file,
    import_ed25519_privatekey_from_file,
    _generate_and_write_ecdsa_keypair,
    generate_and_write_ecdsa_keypair,
    generate_and_write_ecdsa_keypair_with_prompt,
    generate_and_write_unencrypted_ecdsa_keypair,
    import_ecdsa_publickey_from_file,
    import_ecdsa_privatekey_from_file,
    import_publickeys_from_file,
    import_privatekey_from_file)



class TestInterfaceFunctions(unittest.TestCase):
  @classmethod
  def setUpClass(cls):
    cls.test_data_dir = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "data")

    cls.path_rsa = os.path.join(
        cls.test_data_dir, "keystore", "rsa_key")
    cls.path_ed25519 = os.path.join(
        cls.test_data_dir, "keystore", "ed25519_key")
    cls.path_ecdsa = os.path.join(
        cls.test_data_dir, "keystore", "ecdsa_key")
    cls.path_no_key = os.path.join(
        cls.test_data_dir, "keystore", "no_key")

    cls.orig_cwd = os.getcwd()

  def setUp(self):
    self.tmp_dir = tempfile.mkdtemp(dir=self.orig_cwd)
    os.chdir(self.tmp_dir)

  def tearDown(self):
    os.chdir(self.orig_cwd)
    shutil.rmtree(self.tmp_dir)


  def test_rsa(self):
    """Test RSA key _generation and import interface functions. """

    # TEST: Generate default keys and import
    # Assert location and format
    fn_default = "default"
    fn_default_ret = _generate_and_write_rsa_keypair(filepath=fn_default)

    pub = import_rsa_publickey_from_file(fn_default + ".pub")
    priv = import_rsa_privatekey_from_file(fn_default)

    self.assertEqual(fn_default, fn_default_ret)
    self.assertTrue(RSAKEY_SCHEMA.matches(pub))
    self.assertTrue(PUBLIC_KEY_SCHEMA.matches(pub))
    self.assertTrue(RSAKEY_SCHEMA.matches(priv))
    # NOTE: There is no private key schema, at least check it has a value
    self.assertTrue(priv["keyval"]["private"])


    # TEST: Generate unencrypted keys with empty prompt
    # Assert importable without password
    fn_empty_prompt = "empty_prompt"
    with mock.patch("securesystemslib.interface.get_password", return_value=""):
      _generate_and_write_rsa_keypair(filepath=fn_empty_prompt, prompt=True)
    import_rsa_privatekey_from_file(fn_empty_prompt)


    # TEST: Generate keys with auto-filename, i.e. keyid
    # Assert filename is keyid
    fn_keyid = _generate_and_write_rsa_keypair()
    pub = import_rsa_publickey_from_file(fn_keyid + ".pub")
    priv = import_rsa_privatekey_from_file(fn_keyid)
    self.assertTrue(
        os.path.basename(fn_keyid) == pub["keyid"] == priv["keyid"])


    # TEST: Generate keys with custom bits
    # Assert length
    bits = 4096
    fn_bits = "bits"
    _generate_and_write_rsa_keypair(filepath=fn_bits, bits=bits)

    priv = import_rsa_privatekey_from_file(fn_bits)
    # NOTE: Parse PEM with pyca/cryptography to get the key size property
    obj_bits = load_pem_private_key(
        priv["keyval"]["private"].encode("utf-8"),
        password=None,
        backend=default_backend())

    self.assertEqual(obj_bits.key_size, bits)


    # TEST: Generate two keypairs with encrypted private keys using ...
    pw = "pw"
    fn_encrypted = "encrypted"
    fn_prompt = "prompt"

    # ... a passed pw ...
    _generate_and_write_rsa_keypair(filepath=fn_encrypted, password=pw)
    with mock.patch("securesystemslib.interface.get_password", return_value=pw):
      # ... and a prompted pw.
      _generate_and_write_rsa_keypair(filepath=fn_prompt, prompt=True)

      # Assert that both private keys are importable using the prompted pw ...
      import_rsa_privatekey_from_file(fn_prompt, prompt=True)
      import_rsa_privatekey_from_file(fn_encrypted, prompt=True)

    # ... and the passed pw.
    import_rsa_privatekey_from_file(fn_prompt, password=pw)
    import_rsa_privatekey_from_file(fn_encrypted, password=pw)


    # TEST: Import existing keys with encrypted private key (test regression)
    # Assert format
    pub = import_rsa_publickey_from_file(self.path_rsa + ".pub")
    priv = import_rsa_privatekey_from_file(self.path_rsa, "password")

    self.assertTrue(RSAKEY_SCHEMA.matches(pub))
    self.assertTrue(PUBLIC_KEY_SCHEMA.matches(pub))
    self.assertTrue(RSAKEY_SCHEMA.matches(priv))
    # NOTE: There is no private key schema, at least check it has a value
    self.assertTrue(priv["keyval"]["private"])


    # TEST: Generation errors
    for idx, (kwargs, err_msg) in enumerate([
        # Error on empty password
        ({"password": ""},
          "encryption password must be 1 or more characters long"),
        # Error on 'password' and 'prompt=True'
        ({"password": pw, "prompt": True},
          "passing 'password' and 'prompt=True' is not allowed")]):

      with self.assertRaises(ValueError, msg="(row {})".format(idx)) as ctx:
        _generate_and_write_rsa_keypair(**kwargs)

      self.assertEqual(err_msg, str(ctx.exception),
          "expected: '{}' got: '{}' (row {})".format(
          err_msg, ctx.exception, idx))

    # Error on bad argument format
    for idx, kwargs in enumerate([
        {"bits": 1024}, # Too low
        {"bits": "not-an-int"},
        {"filepath": 123456}, # Not a string
        {"password": 123456}, # Not a string
        {"prompt": "not-a-bool"}]):
      with self.assertRaises(FormatError, msg="(row {})".format(idx)):
        _generate_and_write_rsa_keypair(**kwargs)


    # TEST: Import errors

    # Error public key import
    err_msg = "Invalid public pem"
    with self.assertRaises(Error) as ctx:
      import_rsa_publickey_from_file(fn_default)

    self.assertTrue(err_msg in str(ctx.exception),
        "expected: '{}' got: '{}'".format(err_msg, ctx.exception))

    # Error on private key import...
    for idx, (args, kwargs, err, err_msg) in enumerate([
        # Error on not a private key
        ([fn_default + ".pub"], {}, CryptoError,
          "Could not deserialize key data"),
        # Error on not encrypted
        ([fn_default], {"password": pw}, CryptoError,
          "Password was given but private key is not encrypted"),
        # Error on encrypted but no pw
        ([fn_encrypted], {}, CryptoError,
          "Password was not given but private key is encrypted"),
        # Error on encrypted but empty pw passed
        ([fn_encrypted], {"password": ""}, CryptoError,
          "Password was not given but private key is encrypted"),
        # Error on encrypted but bad pw passed
        ([fn_encrypted], {"password": "bad pw"}, CryptoError,
          "Bad decrypt. Incorrect password?"),
        # Error on pw and prompt
        ([fn_default], {"password": pw, "prompt": True}, ValueError,
          "passing 'password' and 'prompt=True' is not allowed")]):

      with self.assertRaises(err, msg="(row {})".format(idx)) as ctx:
        import_rsa_privatekey_from_file(*args, **kwargs)

      self.assertTrue(err_msg in str(ctx.exception),
          "expected: '{}' got: '{}' (row {})".format(
          err_msg, ctx.exception, idx))

    # Error on encrypted but bad pw prompted
    err_msg = "Password was not given but private key is encrypted"
    with self.assertRaises(CryptoError) as ctx, mock.patch(
        "securesystemslib.interface.get_password", return_value="bad_pw"):
      import_rsa_privatekey_from_file(fn_encrypted)

    self.assertTrue(err_msg in str(ctx.exception),
        "expected: '{}' got: '{}'".format(err_msg, ctx.exception))

    # Error on bad argument format
    for idx, (args, kwargs) in enumerate([
          ([123456], {}), # bad path
          ([fn_default], {"scheme": 123456}), # bad scheme
          ([fn_default], {"scheme": "bad scheme"}) # bad scheme
        ]):
      with self.assertRaises(FormatError, msg="(row {})".format(idx)):
        import_rsa_publickey_from_file(*args, **kwargs)
      with self.assertRaises(FormatError, msg="(row {})".format(idx)):
        import_rsa_privatekey_from_file(*args, **kwargs)

    # bad password
    with self.assertRaises(FormatError):
      import_rsa_privatekey_from_file(fn_default, password=123456)

    # bad prompt
    with self.assertRaises(FormatError):
      import_rsa_privatekey_from_file(fn_default, prompt="not-a-bool")



  def test_ed25519(self):
    """Test ed25519 key _generation and import interface functions. """

    # TEST: Generate default keys and import
    # Assert location and format
    fn_default = "default"
    fn_default_ret = _generate_and_write_ed25519_keypair(filepath=fn_default)

    pub = import_ed25519_publickey_from_file(fn_default + ".pub")
    priv = import_ed25519_privatekey_from_file(fn_default)

    self.assertEqual(fn_default, fn_default_ret)
    self.assertTrue(ED25519KEY_SCHEMA.matches(pub))
    self.assertTrue(PUBLIC_KEY_SCHEMA.matches(pub))
    self.assertTrue(ED25519KEY_SCHEMA.matches(priv))
    # NOTE: There is no private key schema, at least check it has a value
    self.assertTrue(priv["keyval"]["private"])


    # TEST: Generate unencrypted keys with empty prompt
    # Assert importable with empty prompt password and without password
    fn_empty_prompt = "empty_prompt"
    with mock.patch("securesystemslib.interface.get_password", return_value=""):
      _generate_and_write_ed25519_keypair(filepath=fn_empty_prompt)
      import_ed25519_privatekey_from_file(fn_empty_prompt, prompt=True)
    import_ed25519_privatekey_from_file(fn_empty_prompt)


    # TEST: Generate keys with auto-filename, i.e. keyid
    # Assert filename is keyid
    fn_keyid = _generate_and_write_ed25519_keypair()
    pub = import_ed25519_publickey_from_file(fn_keyid + ".pub")
    priv = import_ed25519_privatekey_from_file(fn_keyid)
    self.assertTrue(
        os.path.basename(fn_keyid) == pub["keyid"] == priv["keyid"])


    # TEST: Generate two keypairs with encrypted private keys using ...
    pw = "pw"
    fn_encrypted = "encrypted"
    fn_prompt = "prompt"
    # ... a passed pw ...
    _generate_and_write_ed25519_keypair(filepath=fn_encrypted, password=pw)
    with mock.patch("securesystemslib.interface.get_password", return_value=pw):
      # ... and a prompted pw.
      _generate_and_write_ed25519_keypair(filepath=fn_prompt, prompt=True)

      # Assert that both private keys are importable using the prompted pw ...
      import_ed25519_privatekey_from_file(fn_prompt, prompt=True)
      import_ed25519_privatekey_from_file(fn_encrypted, prompt=True)

    # ... and the passed pw.
    import_ed25519_privatekey_from_file(fn_prompt, password=pw)
    import_ed25519_privatekey_from_file(fn_encrypted, password=pw)


    # TEST: Import existing keys with encrypted private key (test regression)
    # Assert format
    pub = import_ed25519_publickey_from_file(self.path_ed25519 + ".pub")
    priv = import_ed25519_privatekey_from_file(self.path_ed25519, "password")

    self.assertTrue(PUBLIC_KEY_SCHEMA.matches(pub))
    self.assertTrue(ED25519KEY_SCHEMA.matches(pub))
    self.assertTrue(ED25519KEY_SCHEMA.matches(priv))
    # NOTE: There is no private key schema, at least check it has a value
    self.assertTrue(priv["keyval"]["private"])


    # TEST: Unexpected behavior
    # FIXME: Should 'import_ed25519_publickey_from_file' be able to import a
    # a non-encrypted ed25519 private key? I think it should not, but it is:
    priv = import_ed25519_publickey_from_file(fn_default)
    self.assertTrue(ED25519KEY_SCHEMA.matches(priv))
    self.assertTrue(priv["keyval"]["private"])

    # FIXME: Should 'import_ed25519_privatekey_from_file' be able to import a
    # an ed25519 public key? I think it should not, but it is:
    pub = import_ed25519_privatekey_from_file(fn_default + ".pub")
    self.assertTrue(PUBLIC_KEY_SCHEMA.matches(pub))


    # TEST: Generation errors
    for idx, (kwargs, err_msg) in enumerate([
        # Error on empty password
        ({"password": ""},
          "encryption password must be 1 or more characters long"),
        # Error on 'password' and 'prompt=True'
        ({"password": pw, "prompt": True},
          "passing 'password' and 'prompt=True' is not allowed")]):

      with self.assertRaises(ValueError, msg="(row {})".format(idx)) as ctx:
        _generate_and_write_ed25519_keypair(**kwargs)

      self.assertEqual(err_msg, str(ctx.exception),
          "expected: '{}' got: '{}' (row {})".format(
          err_msg, ctx.exception, idx))

    # Error on bad argument format
    for idx, kwargs in enumerate([
        {"filepath": 123456}, # Not a string
        {"password": 123456}, # Not a string
        {"prompt": "not-a-bool"}]):
      with self.assertRaises(FormatError, msg="(row {})".format(idx)):
        _generate_and_write_ed25519_keypair(**kwargs)


    # TEST: Import errors
    # Error on public key import...
    for idx, (fn, err_msg) in enumerate([
        # Error on invalid json (custom key format)
        (fn_encrypted, "Cannot deserialize to a Python object"),
        # Error on invalid custom key format
        (self.path_no_key, "Missing key" ),
        # Error on invalid key type
        (self.path_ecdsa + ".pub", "Invalid key type loaded")]):
      with self.assertRaises(Error, msg="(row {})".format(idx)) as ctx:
        import_ed25519_publickey_from_file(fn)

      self.assertTrue(err_msg in str(ctx.exception),
          "expected: '{}' got: '{}' (row {})".format(
          err_msg, ctx.exception, idx))

    # Error on private key import...
    for idx, (args, kwargs, err, err_msg) in enumerate([
        # Error on not an ed25519 private key
        ([self.path_ecdsa], {}, CryptoError,
          "Malformed Ed25519 key JSON, possibly due to encryption, "
          "but no password provided?"),
        # Error on not encrypted
        ([fn_default], {"password": pw}, CryptoError,
          "Invalid encrypted file."),
        # Error on encrypted but no pw
        ([fn_encrypted], {}, CryptoError,
          "Malformed Ed25519 key JSON, possibly due to encryption, "
          "but no password provided?"),
        # Error on encrypted but empty pw
        ([fn_encrypted], {"password": ""}, CryptoError,
          "Decryption failed."),
        # Error on encrypted but bad pw passed
        ([fn_encrypted], {"password": "bad pw"}, CryptoError,
          "Decryption failed."),
        # Error on pw and prompt
        ([fn_default], {"password": pw, "prompt": True}, ValueError,
          "passing 'password' and 'prompt=True' is not allowed")]):

      with self.assertRaises(err, msg="(row {})".format(idx)) as ctx:
        import_ed25519_privatekey_from_file(*args, **kwargs)

      self.assertTrue(err_msg in str(ctx.exception),
          "expected: '{}' got: '{}' (row {})".format(
          err_msg, ctx.exception, idx))


    # Error on encrypted but bad pw prompted
    err_msg = ("Malformed Ed25519 key JSON, possibly due to encryption, "
        "but no password provided?")
    with self.assertRaises(CryptoError) as ctx, mock.patch(
        "securesystemslib.interface.get_password", return_value="bad_pw"):
      import_ed25519_privatekey_from_file(fn_encrypted)

    self.assertTrue(err_msg in str(ctx.exception),
        "expected: '{}' got: '{}'".format(err_msg, ctx.exception))


    # Error on bad path format
    with self.assertRaises(FormatError):
      import_ed25519_publickey_from_file(123456)
    with self.assertRaises(FormatError):
      import_ed25519_privatekey_from_file(123456)

    # Error on bad password format
    with self.assertRaises(FormatError):
      import_ed25519_privatekey_from_file(fn_default, password=123456)

    # Error on bad prompt format
    with self.assertRaises(FormatError):
      import_ed25519_privatekey_from_file(fn_default, prompt="not-a-bool")


  def test_ecdsa(self):
    """Test ecdsa key _generation and import interface functions. """
    # TEST: Generate default keys and import
    # Assert location and format
    fn_default = "default"
    fn_default_ret = _generate_and_write_ecdsa_keypair(filepath=fn_default)

    pub = import_ecdsa_publickey_from_file(fn_default + ".pub")
    priv = import_ecdsa_privatekey_from_file(fn_default)

    self.assertEqual(fn_default, fn_default_ret)
    self.assertTrue(ECDSAKEY_SCHEMA.matches(pub))
    self.assertTrue(PUBLIC_KEY_SCHEMA.matches(pub))
    self.assertTrue(ECDSAKEY_SCHEMA.matches(priv))
    # NOTE: There is no private key schema, at least check it has a value
    self.assertTrue(priv["keyval"]["private"])

    # TEST: Generate unencrypted keys with empty prompt
    # Assert importable with empty prompt password and without password
    fn_empty_prompt = "empty_prompt"
    with mock.patch("securesystemslib.interface.get_password", return_value=""):
      _generate_and_write_ecdsa_keypair(filepath=fn_empty_prompt)
      import_ecdsa_privatekey_from_file(fn_empty_prompt, prompt=True)
    import_ecdsa_privatekey_from_file(fn_empty_prompt)


    # TEST: Generate keys with auto-filename, i.e. keyid
    # Assert filename is keyid
    fn_keyid = _generate_and_write_ecdsa_keypair()
    pub = import_ecdsa_publickey_from_file(fn_keyid + ".pub")
    priv = import_ecdsa_privatekey_from_file(fn_keyid)
    self.assertTrue(
        os.path.basename(fn_keyid) == pub["keyid"] == priv["keyid"])


    # TEST: Generate two key pairs with encrypted private keys using ...
    pw = "pw"
    fn_encrypted = "encrypted"
    fn_prompt = "prompt"
    # ...  a passed pw ...
    _generate_and_write_ecdsa_keypair(filepath=fn_encrypted, password=pw)
    with mock.patch("securesystemslib.interface.get_password", return_value=pw):
      # ... and a prompted pw.
      _generate_and_write_ecdsa_keypair(filepath=fn_prompt, prompt=True)

      # Assert that both private keys are importable using the prompted pw ...
      import_ecdsa_privatekey_from_file(fn_prompt, prompt=True)
      import_ecdsa_privatekey_from_file(fn_encrypted, prompt=True)

    # ... and the passed pw.
    import_ecdsa_privatekey_from_file(fn_prompt, password=pw)
    import_ecdsa_privatekey_from_file(fn_encrypted, password=pw)


    # TEST: Import existing keys with encrypted private key (test regression)
    # Assert format
    pub = import_ecdsa_publickey_from_file(self.path_ecdsa + ".pub")
    priv = import_ecdsa_privatekey_from_file(self.path_ecdsa, "password")

    self.assertTrue(ECDSAKEY_SCHEMA.matches(pub))
    self.assertTrue(PUBLIC_KEY_SCHEMA.matches(pub))
    self.assertTrue(ECDSAKEY_SCHEMA.matches(priv))
    # NOTE: There is no private key schema, at least check it has a value
    self.assertTrue(priv["keyval"]["private"])


    # FIXME: Should 'import_ecdsa_publickey_from_file' be able to import a
    # an ed25519 public key? I think it should not, but it is:
    import_ecdsa_publickey_from_file(self.path_ed25519 + ".pub")
    self.assertTrue(ECDSAKEY_SCHEMA.matches(pub))


    # TEST: Generation errors
    for idx, (kwargs, err_msg) in enumerate([
        # Error on empty password
        ({"password": ""},
          "encryption password must be 1 or more characters long"),
        # Error on 'password' and 'prompt=True'
        ({"password": pw, "prompt": True},
          "passing 'password' and 'prompt=True' is not allowed")]):

      with self.assertRaises(ValueError, msg="(row {})".format(idx)) as ctx:
        _generate_and_write_ecdsa_keypair(**kwargs)

      self.assertEqual(err_msg, str(ctx.exception),
          "expected: '{}' got: '{}' (row {})".format(
          err_msg, ctx.exception, idx))

    # Error on bad argument format
    for idx, kwargs in enumerate([
        {"filepath": 123456}, # Not a string
        {"password": 123456}, # Not a string
        {"prompt": "not-a-bool"}]):
      with self.assertRaises(FormatError, msg="(row {})".format(idx)):
        _generate_and_write_ecdsa_keypair(**kwargs)


    # TEST: Import errors

    # Error on public key import...
    for idx, (fn, err_msg) in enumerate([
        # Error on invalid json (custom key format)
        (fn_encrypted, "Cannot deserialize to a Python object"),
        # Error on invalid custom key format
        (self.path_no_key, "Missing key")]):
      with self.assertRaises(Error, msg="(row {})".format(idx)) as ctx:
        import_ecdsa_publickey_from_file(fn)

      self.assertTrue(err_msg in str(ctx.exception),
          "expected: '{}' got: '{}' (row {})".format(
          err_msg, ctx.exception, idx))


    # Error on private key import...
    for idx, (args, kwargs, err, err_msg) in enumerate([
        # Error on not an ecdsa private key
        ([self.path_ed25519], {}, Error,
          "Cannot deserialize to a Python object"),
        # Error on not encrypted
        ([fn_default], {"password": pw}, CryptoError,
          "Invalid encrypted file."),
        # Error on encrypted but no pw
        ([fn_encrypted], {}, Error,
          "Cannot deserialize to a Python object"),
        # Error on encrypted but empty pw
        ([fn_encrypted], {"password": ""}, CryptoError,
          "Decryption failed."),
        # Error on encrypted but bad pw passed
        ([fn_encrypted], {"password": "bad pw"}, CryptoError,
          "Decryption failed."),
        # Error on pw and prompt
        ([fn_default], {"password": pw, "prompt": True}, ValueError,
          "passing 'password' and 'prompt=True' is not allowed")]):

      with self.assertRaises(err, msg="(row {})".format(idx)) as ctx:
        import_ecdsa_privatekey_from_file(*args, **kwargs)

      self.assertTrue(err_msg in str(ctx.exception),
          "expected: '{}' got: '{}' (row {})".format(
          err_msg, ctx.exception, idx))

    # Error on encrypted but bad pw prompted
    err_msg = ("Decryption failed")
    with self.assertRaises(CryptoError) as ctx, mock.patch(
        "securesystemslib.interface.get_password", return_value="bad_pw"):
      import_ecdsa_privatekey_from_file(fn_encrypted, prompt=True)

    self.assertTrue(err_msg in str(ctx.exception),
        "expected: '{}' got: '{}'".format(err_msg, ctx.exception))


    # Error on bad path format
    with self.assertRaises(FormatError):
      import_ecdsa_publickey_from_file(123456)
    with self.assertRaises(FormatError):
      import_ecdsa_privatekey_from_file(123456)

    # Error on bad password format
    with self.assertRaises(FormatError): # bad password
      import_ecdsa_privatekey_from_file(fn_default, password=123456)

    # Error on bad prompt format
    with self.assertRaises(FormatError):
      import_ecdsa_privatekey_from_file(fn_default, prompt="not-a-bool")



  def test_generate_keypair_wrappers(self):
    """Basic tests for thin wrappers around _generate_and_write_*_keypair.
    See 'test_rsa', 'test_ed25519' and 'test_ecdsa' for more thorough key
    generation tests for each key type.

    """
    key_pw = "pw"
    for idx, (gen, gen_prompt, gen_plain, import_priv, schema) in enumerate([
        (
          generate_and_write_rsa_keypair,
          generate_and_write_rsa_keypair_with_prompt,
          generate_and_write_unencrypted_rsa_keypair,
          import_rsa_privatekey_from_file,
          RSAKEY_SCHEMA
        ),
        (
          generate_and_write_ed25519_keypair,
          generate_and_write_ed25519_keypair_with_prompt,
          generate_and_write_unencrypted_ed25519_keypair,
          import_ed25519_privatekey_from_file,
          ED25519KEY_SCHEMA
        ),
        (
          generate_and_write_ecdsa_keypair,
          generate_and_write_ecdsa_keypair_with_prompt,
          generate_and_write_unencrypted_ecdsa_keypair,
          import_ecdsa_privatekey_from_file,
          ECDSAKEY_SCHEMA)]):

      assert_msg = "(row {})".format(idx)
      # Test generate_and_write_*_keypair creates an encrypted private key
      fn_encrypted = gen(key_pw)
      priv = import_priv(fn_encrypted, key_pw)
      self.assertTrue(schema.matches(priv), assert_msg)

      # Test generate_and_write_*_keypair errors if password is None or empty
      with self.assertRaises(FormatError, msg=assert_msg):
        fn_encrypted = gen(None)
      with self.assertRaises(ValueError, msg=assert_msg):
        fn_encrypted = gen("")

      # Test generate_and_write_*_keypair_with_prompt creates encrypted private
      # key
      with mock.patch(
          "securesystemslib.interface.get_password", return_value=key_pw):
        fn_prompt = gen_prompt()
      priv = import_priv(fn_prompt, key_pw)
      self.assertTrue(schema.matches(priv), assert_msg)

      # Test generate_and_write_*_keypair_with_prompt creates unencrypted
      # private key if no password is entered
      with mock.patch(
          "securesystemslib.interface.get_password", return_value=""):
        fn_empty_prompt = gen_prompt()
      priv = import_priv(fn_empty_prompt)
      self.assertTrue(schema.matches(priv), assert_msg)

      # Test generate_and_write_unencrypted_*_keypair doesn't encrypt
      fn_unencrypted = gen_plain()
      priv = import_priv(fn_unencrypted)
      self.assertTrue(schema.matches(priv), assert_msg)



  def test_import_publickeys_from_file(self):
    """Test import multiple public keys with different types. """

    # Successfully import key dict with one key per supported key type
    key_dict = import_publickeys_from_file([
        self.path_rsa + ".pub",
        self.path_ed25519  + ".pub",
        self.path_ecdsa  + ".pub"],
        [KEY_TYPE_RSA, KEY_TYPE_ED25519, KEY_TYPE_ECDSA])

    ANY_PUBKEY_DICT_SCHEMA.check_match(key_dict)
    self.assertListEqual(
        sorted([key["keytype"] for key in key_dict.values()]),
        sorted([KEY_TYPE_RSA, KEY_TYPE_ED25519, KEY_TYPE_ECDSA])
      )

    # Successfully import default rsa key
    key_dict = import_publickeys_from_file([self.path_rsa + ".pub"])
    ANY_PUBKEY_DICT_SCHEMA.check_match(key_dict)
    RSAKEY_SCHEMA.check_match(
        list(key_dict.values()).pop())

    # Bad default rsa key type for ed25519
    with self.assertRaises(Error):
      import_publickeys_from_file([self.path_ed25519 + ".pub"])

    # Bad ed25519 key type for rsa key
    with self.assertRaises(Error):
      import_publickeys_from_file(
          [self.path_rsa + ".pub"], [KEY_TYPE_ED25519])

    # Unsupported key type
    with self.assertRaises(FormatError):
      import_publickeys_from_file(
          [self.path_ed25519 + ".pub"], ["KEY_TYPE_UNSUPPORTED"])

    # Mismatching arguments lists lenghts
    with self.assertRaises(FormatError):
      import_publickeys_from_file(
          [self.path_rsa + ".pub", self.path_ed25519 + ".pub"],
          [KEY_TYPE_ED25519])


  def test_import_privatekey_from_file(self):
    """Test generic private key import function. """

    pw = "password"
    for idx, (path, key_type, key_schema) in enumerate([
        (self.path_rsa, None, RSAKEY_SCHEMA), # default key type
        (self.path_rsa, KEY_TYPE_RSA, RSAKEY_SCHEMA),
        (self.path_ed25519, KEY_TYPE_ED25519, ED25519KEY_SCHEMA),
        (self.path_ecdsa, KEY_TYPE_ECDSA, ECDSAKEY_SCHEMA)]):

      # Successfully import key per supported type, with ...
      # ... passed password
      key = import_privatekey_from_file(path, key_type=key_type, password=pw)
      self.assertTrue(key_schema.matches(key), "(row {})".format(idx))

      # ... entered password on mock-prompt
      with mock.patch("securesystemslib.interface.get_password", return_value=pw):
        key = import_privatekey_from_file(path, key_type=key_type, prompt=True)
      self.assertTrue(key_schema.matches(key), "(row {})".format(idx))

    # Error on wrong key for default key type
    with self.assertRaises(Error):
      import_privatekey_from_file(self.path_ed25519, password=pw)

    # Error on unsupported key type
    with self.assertRaises(FormatError):
      import_privatekey_from_file(
          self.path_rsa, key_type="KEY_TYPE_UNSUPPORTED", password=pw)



# Run the test cases.
if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
"""
<Program Name>
  process.py

<Author>
  Trishank Karthik Kuppusamy <trishank.kuppusamy@datadoghq.com>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  September 25, 2018

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provide a common interface for Python's subprocess module to:

  - require the Py3 subprocess backport `subprocess32` on Python2,
  - namespace subprocess constants (DEVNULL, PIPE) and
  - provide a custom `subprocess.run` wrapper
  - provide a special `run_duplicate_streams` function

"""
import os
import sys
import io
import tempfile
import logging
import time
import shlex
import six

if six.PY2:
  import subprocess32 as subprocess # pragma: no cover pylint: disable=import-error
else: # pragma: no cover
  import subprocess

from securesystemslib import formats
from securesystemslib import settings

DEVNULL = subprocess.DEVNULL
PIPE = subprocess.PIPE

log = logging.getLogger(__name__)

def _default_timeout():
  """Helper to use securesystemslib.settings.SUBPROCESS_TIMEOUT as default
  argument, and still be able to modify it after the function definitions are
  evaluated. """
  return settings.SUBPROCESS_TIMEOUT



def run(cmd, check=True, timeout=_default_timeout(), **kwargs):
  """
  <Purpose>
    Provide wrapper for `subprocess.run` (see
    https://github.com/python/cpython/blob/3.5/Lib/subprocess.py#L352-L399)
    where:

    * `timeout` has a default (securesystemslib.settings.SUBPROCESS_TIMEOUT),
    * `check` is `True` by default,
    * there is only one positional argument, i.e. `cmd` that can be either
      a str (will be split with shlex) or a list of str and
    * instead of raising a ValueError if both `input` and `stdin` are passed,
      `stdin` is ignored.


  <Arguments>
    cmd:
            The command and its arguments. (list of str, or str)
            Splits a string specifying a command and its argument into a list
            of substrings, if necessary.

    check: (default True)
            "If check is true, and the process exits with a non-zero exit code,
            a CalledProcessError exception will be raised. Attributes of that
            exception hold the arguments, the exit code, and stdout and stderr
            if they were captured."

    timeout: (default see securesystemslib.settings.SUBPROCESS_TIMEOUT)
            "The timeout argument is passed to Popen.communicate(). If the
            timeout expires, the child process will be killed and waited for.
            The TimeoutExpired exception will be re-raised after the child
            process has terminated."

    **kwargs:
            See subprocess.run and Frequently Used Arguments to Popen
            constructor for available kwargs.
            https://docs.python.org/3.5/library/subprocess.html#subprocess.run
            https://docs.python.org/3.5/library/subprocess.html#frequently-used-arguments

  <Exceptions>
    securesystemslib.exceptions.FormatError:
            If the `cmd` is a list and does not match
            securesystemslib.formats.LIST_OF_ANY_STRING_SCHEMA.

    OSError:
            If the given command is not present or non-executable.

    subprocess.TimeoutExpired:
            If the process does not terminate after timeout seconds. Default
            is `settings.SUBPROCESS_TIMEOUT`

  <Side Effects>
    The side effects of executing the given command in this environment.

  <Returns>
    A subprocess.CompletedProcess instance.

  """
  # Make list of command passed as string for convenience
  if isinstance(cmd, six.string_types):
    cmd = shlex.split(cmd)
  else:
    formats.LIST_OF_ANY_STRING_SCHEMA.check_match(cmd)

  # NOTE: The CPython implementation would raise a ValueError here, we just
  # don't pass on `stdin` if the user passes `input` and `stdin`
  # https://github.com/python/cpython/blob/3.5/Lib/subprocess.py#L378-L381
  if kwargs.get("input") is not None and "stdin" in kwargs:
    log.debug("stdin and input arguments may not both be used. "
        "Ignoring passed stdin: " + str(kwargs["stdin"]))
    del kwargs["stdin"]

  return subprocess.run(cmd, check=check, timeout=timeout, **kwargs)




def run_duplicate_streams(cmd, timeout=_default_timeout()):
  """
  <Purpose>
    Provide a function that executes a command in a subprocess and, upon
    termination, returns its exit code and the contents of what was printed to
    its standard streams.

    * Might behave unexpectedly with interactive commands.
    * Might not duplicate output in real time, if the command buffers it (see
      e.g. `print("foo")` vs. `print("foo", flush=True)` in Python 3).

  <Arguments>
    cmd:
            The command and its arguments. (list of str, or str)
            Splits a string specifying a command and its argument into a list
            of substrings, if necessary.

    timeout: (default see settings.SUBPROCESS_TIMEOUT)
            If the timeout expires, the child process will be killed and waited
            for and then subprocess.TimeoutExpired will be raised.

  <Exceptions>
    securesystemslib.exceptions.FormatError:
            If the `cmd` is a list and does not match
            securesystemslib.formats.LIST_OF_ANY_STRING_SCHEMA.

    OSError:
            If the given command is not present or non-executable.

    subprocess.TimeoutExpired:
            If the process does not terminate after timeout seconds. Default
            is `settings.SUBPROCESS_TIMEOUT`

  <Side Effects>
    The side effects of executing the given command in this environment.

  <Returns>
    A tuple of command's exit code, standard output and standard error
    contents.

  """
  if isinstance(cmd, six.string_types):
    cmd = shlex.split(cmd)
  else:
    formats.LIST_OF_ANY_STRING_SCHEMA.check_match(cmd)

  # Use temporary files as targets for child process standard stream redirects
  # They seem to work better (i.e. do not hang) than pipes, when using
  # interactive commands like `vi`.
  stdout_fd, stdout_name = tempfile.mkstemp()
  stderr_fd, stderr_name = tempfile.mkstemp()
  try:
    with io.open(stdout_name, "r") as stdout_reader, \
        os.fdopen(stdout_fd, "w") as stdout_writer, \
        io.open(stderr_name, "r") as stderr_reader, \
        os.fdopen(stderr_fd, "w") as stderr_writer:

      # Store stream results in mutable dict to update it inside nested helper
      _std = {"out": "", "err": ""}
      def _duplicate_streams():
        """Helper to read from child process standard streams, write their
        contents to parent process standard streams, and build up return values
        for outer function.
        """
        # Read until EOF but at most `io.DEFAULT_BUFFER_SIZE` bytes per call.
        # Reading and writing in reasonably sized chunks prevents us from
        # subverting a timeout, due to being busy for too long or indefinitely.
        stdout_part = stdout_reader.read(io.DEFAULT_BUFFER_SIZE)
        stderr_part = stderr_reader.read(io.DEFAULT_BUFFER_SIZE)
        sys.stdout.write(stdout_part)
        sys.stderr.write(stderr_part)
        sys.stdout.flush()
        sys.stderr.flush()
        _std["out"] += stdout_part
        _std["err"] += stderr_part

      # Start child process, writing its standard streams to temporary files
      proc = subprocess.Popen(cmd, stdout=stdout_writer,
          stderr=stderr_writer, universal_newlines=True)
      proc_start_time = time.time()

      # Duplicate streams until the process exits (or times out)
      while proc.poll() is None:
        # Time out as Python's `subprocess` would do it
        if (timeout is not None and
            time.time() > proc_start_time + timeout):
          proc.kill()
          proc.wait()
          raise subprocess.TimeoutExpired(cmd, timeout)

        _duplicate_streams()

      # Read/write once more to grab everything that the process wrote between
      # our last read in the loop and exiting, i.e. breaking the loop.
      _duplicate_streams()

  finally:
    # The work is done or was interrupted, the temp files can be removed
    os.remove(stdout_name)
    os.remove(stderr_name)

  # Return process exit code and captured streams
  return proc.poll(), _std["out"], _std["err"]



"""
We copy some functions from the Python 3.3.0 ssl module.

http://hg.python.org/releasing/3.3.0/file/1465cbbc8f64/Lib/ssl.py
"""


import re


class CertificateError(ValueError):
    pass


def _dnsname_to_pat(dn):
    pats = []
    for frag in dn.split(r'.'):
        if frag == '*':
            # When '*' is a fragment by itself, it matches a non-empty dotless
            # fragment.
            pats.append('[^.]+')
        else:
            # Otherwise, '*' matches any dotless fragment.
            frag = re.escape(frag)
            pats.append(frag.replace(r'\*', '[^.]*'))
    return re.compile(r'\A' + r'\.'.join(pats) + r'\Z', re.IGNORECASE)


def match_hostname(cert, hostname):
    """Verify that *cert* (in decoded format as returned by
    SSLSocket.getpeercert()) matches the *hostname*.  RFC 2818 rules
    are mostly followed, but IP addresses are not accepted for *hostname*.

    CertificateError is raised on failure. On success, the function
    returns nothing.
    """
    if not cert:
        raise ValueError("empty or no certificate")
    dnsnames = []
    san = cert.get('subjectAltName', ())
    for key, value in san:
        if key == 'DNS':
            if _dnsname_to_pat(value).match(hostname):
                return
            dnsnames.append(value)
    if not dnsnames:
        # The subject is only checked when there is no dNSName entry
        # in subjectAltName
        for sub in cert.get('subject', ()):
            for key, value in sub:
                # XXX according to RFC 2818, the most specific Common Name
                # must be used.
                if key == 'commonName':
                    if _dnsname_to_pat(value).match(hostname):
                        return
                    dnsnames.append(value)
    if len(dnsnames) > 1:
        raise CertificateError("hostname %r "
            "doesn't match either of %s"
            % (hostname, ', '.join(map(repr, dnsnames))))
    elif len(dnsnames) == 1:
        raise CertificateError("hostname %r "
            "doesn't match %r"
            % (hostname, dnsnames[0]))
    else:
        raise CertificateError("no appropriate commonName or "
            "subjectAltName fields were found")






"""
<Program Name>
  storage.py

<Author>
  Joshua Lock <jlock@vmware.com>

<Started>
  April 9, 2020

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides an interface for filesystem interactions, StorageBackendInterface.
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import abc
import errno
import logging
import os
import shutil

import six

from securesystemslib import exceptions

logger = logging.getLogger(__name__)

if six.PY2:
  FileNotFoundError = OSError # pragma: no cover



class StorageBackendInterface():
  """
  <Purpose>
  Defines an interface for abstract storage operations which can be implemented
  for a variety of storage solutions, such as remote and local filesystems.
  """

  __metaclass__ = abc.ABCMeta


  @abc.abstractmethod
  def get(self, filepath):
    """
    <Purpose>
      A context manager for 'with' statements that is used for retrieving files
      from a storage backend and cleans up the files upon exit.

        with storage_backend.get('/path/to/file') as file_object:
          # operations
        # file is now closed

    <Arguments>
      filepath:
        The full path of the file to be retrieved.

    <Exceptions>
      securesystemslib.exceptions.StorageError, if the file does not exist or is
      no accessible.

    <Returns>
      A ContextManager object that emits a file-like object for the file at
      'filepath'.
    """
    raise NotImplementedError # pragma: no cover


  @abc.abstractmethod
  def put(self, fileobj, filepath):
    """
    <Purpose>
      Store a file-like object in the storage backend.
      The file-like object is read from the beginning, not its current
      offset (if any).

    <Arguments>
      fileobj:
        The file-like object to be stored.

      filepath:
        The full path to the location where 'fileobj' will be stored.

    <Exceptions>
      securesystemslib.exceptions.StorageError, if the file can not be stored.

    <Returns>
      None
    """
    raise NotImplementedError # pragma: no cover


  @abc.abstractmethod
  def remove(self, filepath):
    """
    <Purpose>
      Remove the file at 'filepath' from the storage.

    <Arguments>
      filepath:
        The full path to the file.

    <Exceptions>
      securesystemslib.exceptions.StorageError, if the file can not be removed.

    <Returns>
      None
    """
    raise NotImplementedError # pragma: no cover


  @abc.abstractmethod
  def getsize(self, filepath):
    """
    <Purpose>
      Retrieve the size, in bytes, of the file at 'filepath'.

    <Arguments>
      filepath:
        The full path to the file.

    <Exceptions>
      securesystemslib.exceptions.StorageError, if the file does not exist or is
      not accessible.

    <Returns>
      The size in bytes of the file at 'filepath'.
    """
    raise NotImplementedError # pragma: no cover


  @abc.abstractmethod
  def create_folder(self, filepath):
    """
    <Purpose>
      Create a folder at filepath and ensure all intermediate components of the
      path exist.
      Passing an empty string for filepath does nothing and does not raise an
      exception.

    <Arguments>
      filepath:
        The full path of the folder to be created.

    <Exceptions>
      securesystemslib.exceptions.StorageError, if the folder can not be
      created.

    <Returns>
      None
    """
    raise NotImplementedError # pragma: no cover


  @abc.abstractmethod
  def list_folder(self, filepath):
    """
    <Purpose>
      List the contents of the folder at 'filepath'.

    <Arguments>
      filepath:
        The full path of the folder to be listed.

    <Exceptions>
      securesystemslib.exceptions.StorageError, if the file does not exist or is
      not accessible.

    <Returns>
      A list containing the names of the files in the folder. May be an empty
      list.
    """
    raise NotImplementedError # pragma: no cover





class FilesystemBackend(StorageBackendInterface):
  """
  <Purpose>
    A concrete implementation of StorageBackendInterface which interacts with
    local filesystems using Python standard library functions.
  """

  # As FilesystemBackend is effectively a stateless wrapper around various
  # standard library operations, we only ever need a single instance of it.
  # That single instance is safe to be (re-)used by all callers. Therefore
  # implement the singleton pattern to avoid uneccesarily creating multiple
  # objects.
  _instance = None

  def __new__(cls, *args, **kwargs):
    if cls._instance is None:
      cls._instance = object.__new__(cls, *args, **kwargs)
    return cls._instance



  class GetFile(object):
    # Implementing get() as a function with the @contextmanager decorator
    # doesn't allow us to cleanly capture exceptions thrown by the underlying
    # implementation and bubble up our generic
    # securesystemslib.exceptions.StorageError, therefore we implement get as
    # a class and also assign the class to the 'get' attribute of the parent
    # FilesystemBackend class.

    def __init__(self, filepath):
      self.filepath = filepath


    def __enter__(self):
      try:
        self.file_object = open(self.filepath, 'rb')
        return self.file_object
      except (FileNotFoundError, IOError):
        raise exceptions.StorageError(
            "Can't open %s" % self.filepath)


    def __exit__(self, exc_type, exc_val, traceback):
      self.file_object.close()



  # Map our class ContextManager implementation to the function expected of the
  # securesystemslib.storage.StorageBackendInterface.get definition
  get = GetFile


  def put(self, fileobj, filepath):
    # If we are passed an open file, seek to the beginning such that we are
    # copying the entire contents
    if not fileobj.closed:
      fileobj.seek(0)

    try:
      with open(filepath, 'wb') as destination_file:
        shutil.copyfileobj(fileobj, destination_file)
        # Force the destination file to be written to disk from Python's internal
        # and the operating system's buffers.  os.fsync() should follow flush().
        destination_file.flush()
        os.fsync(destination_file.fileno())
    except (OSError, IOError):
      raise exceptions.StorageError(
          "Can't write file %s" % filepath)


  def remove(self, filepath):
    try:
      os.remove(filepath)
    except (FileNotFoundError, PermissionError, OSError):  # pragma: no cover
      raise exceptions.StorageError(
          "Can't remove file %s" % filepath)


  def getsize(self, filepath):
    try:
      return os.path.getsize(filepath)
    except OSError:
      raise exceptions.StorageError(
          "Can't access file %s" % filepath)


  def create_folder(self, filepath):
    try:
      os.makedirs(filepath)
    except OSError as e:
      # 'OSError' raised if the leaf directory already exists or cannot be
      # created. Check for case where 'filepath' has already been created and
      # silently ignore.
      if e.errno == errno.EEXIST:
        pass
      elif e.errno == errno.ENOENT and not filepath:
        raise exceptions.StorageError(
            "Can't create a folder with an empty filepath!")
      else:
        raise exceptions.StorageError(
            "Can't create folder at %s" % filepath)


  def list_folder(self, filepath):
    try:
      return os.listdir(filepath)
    except FileNotFoundError:
      raise exceptions.StorageError(
          "Can't list folder at %s" % filepath)



#!/usr/bin/env python

"""
<Program Name>
  test_rsa_keys.py

<Author>
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  June 3, 2015.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Test cases for 'rsa_keys.py'.
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import unittest

import securesystemslib.exceptions
import securesystemslib.formats
import securesystemslib.keys
import securesystemslib.rsa_keys

from cryptography.hazmat.primitives import hashes

public_rsa, private_rsa = securesystemslib.rsa_keys.generate_rsa_public_and_private()
FORMAT_ERROR_MSG = 'securesystemslib.exceptions.FormatError raised.  Check object\'s format.'


class TestRSA_keys(unittest.TestCase):
  def setUp(self):
    pass


  def test_generate_rsa_public_and_private(self):
    pub, priv = securesystemslib.rsa_keys.generate_rsa_public_and_private()

    # Check format of 'pub' and 'priv'.
    self.assertEqual(None, securesystemslib.formats.PEMRSA_SCHEMA.check_match(pub),
        FORMAT_ERROR_MSG)
    self.assertEqual(None, securesystemslib.formats.PEMRSA_SCHEMA.check_match(priv),
        FORMAT_ERROR_MSG)

    # Check for an invalid "bits" argument.  bits >= 2048.
    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.rsa_keys.generate_rsa_public_and_private, 1024)

    self.assertRaises(securesystemslib.exceptions.FormatError,
        securesystemslib.rsa_keys.generate_rsa_public_and_private, '2048')


  def test_create_rsa_signature(self):
    global private_rsa
    global public_rsa
    data = 'The quick brown fox jumps over the lazy dog'.encode('utf-8')

    for rsa_scheme in securesystemslib.keys.RSA_SIGNATURE_SCHEMES:
      signature, scheme = \
        securesystemslib.rsa_keys.create_rsa_signature(private_rsa, data, rsa_scheme)

      # Verify format of returned values.
      self.assertNotEqual(None, signature)
      self.assertEqual(None,
          securesystemslib.formats.RSA_SCHEME_SCHEMA.check_match(scheme),
          FORMAT_ERROR_MSG)
      self.assertEqual(rsa_scheme, scheme)

      # Check for improperly formatted arguments.
      self.assertRaises(securesystemslib.exceptions.FormatError,
          securesystemslib.rsa_keys.create_rsa_signature, 123, data)

      # Check for an unset private key.
      self.assertRaises(ValueError,
          securesystemslib.rsa_keys.create_rsa_signature, '', data)

      # Check for an invalid PEM.
      self.assertRaises(securesystemslib.exceptions.CryptoError,
          securesystemslib.rsa_keys.create_rsa_signature, '123', data)

      # Check for invalid 'data'.
      self.assertRaises(securesystemslib.exceptions.FormatError,
          securesystemslib.rsa_keys.create_rsa_signature, private_rsa, '')

      self.assertRaises(securesystemslib.exceptions.FormatError,
          securesystemslib.rsa_keys.create_rsa_signature, private_rsa, 123)

      # Check for a missing private key.
      self.assertRaises(securesystemslib.exceptions.CryptoError,
          securesystemslib.rsa_keys.create_rsa_signature, public_rsa, data)

      # Check for a TypeError by attempting to create a signature with an
      # encrypted key.
      encrypted_pem = securesystemslib.rsa_keys.create_rsa_encrypted_pem(
          private_rsa, 'pw')
      self.assertRaises(securesystemslib.exceptions.CryptoError,
          securesystemslib.rsa_keys.create_rsa_signature, encrypted_pem,
          data)


  def test_verify_rsa_signature(self):
    global public_rsa
    global private_rsa
    data = 'The quick brown fox jumps over the lazy dog'.encode('utf-8')

    for rsa_scheme in securesystemslib.keys.RSA_SIGNATURE_SCHEMES:
      signature, scheme = \
        securesystemslib.rsa_keys.create_rsa_signature(private_rsa, data, rsa_scheme)

      valid_signature = \
        securesystemslib.rsa_keys.verify_rsa_signature(signature,
          scheme, public_rsa, data)
      self.assertEqual(True, valid_signature)

      # Check for an invalid public key.
      self.assertRaises(securesystemslib.exceptions.CryptoError,
        securesystemslib.rsa_keys.verify_rsa_signature, signature, scheme,
        private_rsa, data)

      # Check for improperly formatted arguments.
      self.assertRaises(securesystemslib.exceptions.FormatError,
          securesystemslib.rsa_keys.verify_rsa_signature, signature,
          123, public_rsa, data)

      self.assertRaises(securesystemslib.exceptions.FormatError,
          securesystemslib.rsa_keys.verify_rsa_signature, signature,
          scheme, 123, data)

      self.assertRaises(securesystemslib.exceptions.FormatError,
          securesystemslib.rsa_keys.verify_rsa_signature, 123, scheme,
          public_rsa, data)

      self.assertRaises(securesystemslib.exceptions.FormatError,
          securesystemslib.rsa_keys.verify_rsa_signature,
          signature, 'invalid_scheme', public_rsa, data)

      # Check for invalid 'signature' and 'data' arguments.
      self.assertRaises(securesystemslib.exceptions.FormatError,
          securesystemslib.rsa_keys.verify_rsa_signature,
          signature, scheme, public_rsa, 123)

      self.assertEqual(False,
          securesystemslib.rsa_keys.verify_rsa_signature(signature,
          scheme, public_rsa, b'mismatched data'))

      mismatched_signature, scheme = \
        securesystemslib.rsa_keys.create_rsa_signature(private_rsa,
        b'mismatched data')

      self.assertEqual(False,
          securesystemslib.rsa_keys.verify_rsa_signature(mismatched_signature,
          scheme, public_rsa, data))


  def test_create_rsa_encrypted_pem(self):
    global public_rsa
    global private_rsa

    encrypted_pem = \
      securesystemslib.rsa_keys.create_rsa_encrypted_pem(private_rsa,
      'password')
    self.assertTrue(securesystemslib.formats.PEMRSA_SCHEMA.matches(encrypted_pem))

    # Test for invalid private key (via PEM).
    self.assertRaises(securesystemslib.exceptions.CryptoError,
      securesystemslib.rsa_keys.create_rsa_encrypted_pem,
      public_rsa, 'password')

    # Test for invalid arguments.
    self.assertRaises(securesystemslib.exceptions.FormatError,
      securesystemslib.rsa_keys.create_rsa_encrypted_pem,
      public_rsa, 123)

    self.assertRaises(securesystemslib.exceptions.FormatError,
      securesystemslib.rsa_keys.create_rsa_encrypted_pem,
      123, 'password')

    self.assertRaises(ValueError,
      securesystemslib.rsa_keys.create_rsa_encrypted_pem,
      '', 'password')



  def test_create_rsa_public_and_private_from_pem(self):
    global public_rsa
    global private_rsa

    public, private = \
      securesystemslib.rsa_keys.create_rsa_public_and_private_from_pem(
      private_rsa)

    self.assertTrue(securesystemslib.formats.PEMRSA_SCHEMA.matches(public))
    self.assertTrue(securesystemslib.formats.PEMRSA_SCHEMA.matches(private))

    self.assertRaises(securesystemslib.exceptions.CryptoError,
      securesystemslib.rsa_keys.create_rsa_public_and_private_from_pem,
      public_rsa)



  def test_encrypt_key(self):
    global public_rsa
    global private_rsa

    key_object = {'keytype': 'rsa',
        'scheme': 'rsassa-pss-sha256',
        'keyid': '1223',
        'keyval': {'public': public_rsa,
        'private': private_rsa}}

    encrypted_key = securesystemslib.rsa_keys.encrypt_key(key_object,
        'password')
    self.assertTrue(securesystemslib.formats.ENCRYPTEDKEY_SCHEMA.matches(encrypted_key))

    key_object['keyval']['private'] = ''
    self.assertRaises(securesystemslib.exceptions.FormatError,
      securesystemslib.rsa_keys.encrypt_key, key_object, 'password')


  def test_decrypt_key(self):

    # Test for valid arguments.
    global public_rsa
    global private_rsa
    passphrase = 'pw'

    rsa_key = {'keytype': 'rsa',
    'scheme': 'rsassa-pss-sha256',
    'keyid': 'd62247f817883f593cf6c66a5a55292488d457bcf638ae03207dbbba9dbe457d',
    'keyval': {'public': public_rsa, 'private': private_rsa}}

    encrypted_rsa_key = securesystemslib.rsa_keys.encrypt_key(rsa_key,
      passphrase)

    decrypted_rsa_key = securesystemslib.rsa_keys.decrypt_key(encrypted_rsa_key,
      passphrase)

    # Test for invalid arguments.
    self.assertRaises(securesystemslib.exceptions.CryptoError,
      securesystemslib.rsa_keys.decrypt_key, 'bad', passphrase)

    # Test for invalid encrypted content (i.e., invalid hmac and ciphertext.)
    encryption_delimiter = securesystemslib.rsa_keys._ENCRYPTION_DELIMITER
    salt, iterations, hmac, iv, ciphertext = \
      encrypted_rsa_key.split(encryption_delimiter)

    # Set an invalid hmac.  The decryption routine sould raise a
    # securesystemslib.exceptions.CryptoError exception because 'hmac' does not
    # match the hmac calculated by the decryption routine.
    bad_hmac = '12345abcd'
    invalid_encrypted_rsa_key = \
      salt + encryption_delimiter + iterations + encryption_delimiter + \
      bad_hmac + encryption_delimiter + iv + encryption_delimiter + ciphertext

    self.assertRaises(securesystemslib.exceptions.CryptoError,
      securesystemslib.rsa_keys.decrypt_key, invalid_encrypted_rsa_key,
      passphrase)

    # Test for invalid 'ciphertext'
    bad_ciphertext = '12345abcde'
    invalid_encrypted_rsa_key = \
      salt + encryption_delimiter + iterations + encryption_delimiter + \
      hmac + encryption_delimiter + iv + encryption_delimiter + bad_ciphertext

    self.assertRaises(securesystemslib.exceptions.CryptoError,
      securesystemslib.rsa_keys.decrypt_key, invalid_encrypted_rsa_key,
      passphrase)



# Run the unit tests.
if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python

"""Test cases for "signer.py". """

import sys
import unittest

import unittest
import securesystemslib.formats
import securesystemslib.keys as KEYS
from securesystemslib.exceptions import FormatError, UnsupportedAlgorithmError

# TODO: Remove case handling when fully dropping support for versions < 3.6
IS_PY_VERSION_SUPPORTED = sys.version_info >= (3, 6)

# Use setUpModule to tell unittest runner to skip this test module gracefully.
def setUpModule():
    if not IS_PY_VERSION_SUPPORTED:
        raise unittest.SkipTest("requires Python 3.6 or higher")

# Since setUpModule is called after imports we need to import conditionally.
if IS_PY_VERSION_SUPPORTED:
    from securesystemslib.signer import Signature, SSlibSigner


class TestSSlibSigner(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.rsakey_dict = KEYS.generate_rsa_key()
        cls.ed25519key_dict = KEYS.generate_ed25519_key()
        cls.ecdsakey_dict = KEYS.generate_ecdsa_key()
        cls.DATA_STR = "SOME DATA REQUIRING AUTHENTICITY."
        cls.DATA = securesystemslib.formats.encode_canonical(
                cls.DATA_STR).encode("utf-8")


    def test_sslib_sign(self):
        dicts = [self.rsakey_dict, self.ecdsakey_dict, self.ed25519key_dict]
        for scheme_dict in dicts:
            # Test generation of signatures.
            sslib_signer = SSlibSigner(scheme_dict)
            sig_obj = sslib_signer.sign(self.DATA)

            # Verify signature
            verified = KEYS.verify_signature(scheme_dict, sig_obj.to_dict(),
                    self.DATA)
            self.assertTrue(verified, "Incorrect signature.")

            # Removing private key from "scheme_dict".
            private = scheme_dict["keyval"]["private"]
            scheme_dict["keyval"]["private"] = ""
            sslib_signer.key_dict = scheme_dict

            with self.assertRaises((ValueError, FormatError)):
                sslib_signer.sign(self.DATA)

            scheme_dict["keyval"]["private"] = private

            # Test for invalid signature scheme.
            valid_scheme = scheme_dict["scheme"]
            scheme_dict["scheme"] = "invalid_scheme"
            sslib_signer = SSlibSigner(scheme_dict)

            with self.assertRaises((UnsupportedAlgorithmError, FormatError)):
                sslib_signer.sign(self.DATA)

            scheme_dict["scheme"] = valid_scheme


    def test_signature_from_to_dict(self):
        signature_dict = {
            "sig": "30460221009342e4566528fcecf6a7a5d53ebacdb1df151e242f55f8775883469cb01dbc6602210086b426cc826709acfa2c3f9214610cb0a832db94bbd266fd7c5939a48064a851",
            "keyid": "11fa391a0ed7a447cbfeb4b2667e286fc248f64d5e6d0eeed2e5e23f97f9f714"
        }
        sig_obj = Signature.from_dict(signature_dict)

        self.assertDictEqual(signature_dict, sig_obj.to_dict())


# Run the unit tests.
if __name__ == "__main__":
    unittest.main()
